import unittest

from fedn.common.exceptions import InvalidParameterError
from fedn.utils.parameters import Parameters


class TestHelpers(unittest.TestCase):

    def test_parameters_invalidkey(self):

        parameters = {
            'serverop': 'adam',
            'learning_rate': 1e-3,
        }
        param = Parameters(parameters)

        parameter_schema = {
            'serveropt': str,
            'learning_rate': float,
        }

        self.assertRaises(InvalidParameterError, param.validate, parameter_schema)

    def test_parameters_valid(self):

        parameters = {
            'serveropt': 'adam',
            'learning_rate': 1e-3,
            'beta1': 0.9,
            'beta2': 0.99,
            'tau': 1e-4,
        }

        param = Parameters(parameters)

        parameter_schema = {
            'serveropt': str,
            'learning_rate': float,
            'beta1': float,
            'beta2': float,
            'tau': float,
        }

        self.assertTrue(param.validate(parameter_schema))

    def test_parameters_invalid(self):

        parameters = {
            'serveropt': 'adam',
            'learning_rate': 1e-3,
            'beta1': 0.9,
            'beta2': 0.99,
            'tau': 1e-4,
        }
        param = Parameters(parameters)

        parameter_schema = {
            'serveropt': str,
            'learning_rate': float,
            'beta1': float,
            'beta2': str,
            'tau': float,
        }

        self.assertRaises(InvalidParameterError, param.validate, parameter_schema)


if __name__ == '__main__':
    unittest.main()



from typing import Tuple

import pymongo
from pymongo.database import Database

from fedn.network.api.shared import statestore_config, network_id

api_version = "v1"

mc = pymongo.MongoClient(**statestore_config["mongo_config"])
mc.server_info()
mdb: Database = mc[network_id]


def is_positive_integer(s):
    return s is not None and s.isdigit() and int(s) > 0


def get_use_typing(headers: object) -> bool:
    skip_typing: str = headers.get("X-Skip-Typing", "false")
    return False if skip_typing.lower() == "true" else True


def get_limit(headers: object) -> int:
    limit: str = headers.get("X-Limit")
    if is_positive_integer(limit):
        return int(limit)
    return 0


def get_reverse(headers: object) -> bool:
    reverse: str = headers.get("X-Reverse")
    if reverse and reverse.lower() == "true":
        return True
    return False


def get_skip(headers: object) -> int:
    skip: str | None = headers.get("X-Skip")
    if is_positive_integer(skip):
        return int(skip)
    return 0


def get_typed_list_headers(
    headers: object,
) -> Tuple[int, int, str, int, bool]:
    sort_key: str = headers.get("X-Sort-Key")
    sort_order: str = headers.get("X-Sort-Order")

    limit: int = get_limit(headers)
    skip: int = get_skip(headers)
    use_typing: bool = get_use_typing(headers)

    if sort_order is not None:
        sort_order = pymongo.ASCENDING if sort_order.lower() == "asc" else pymongo.DESCENDING
    else:
        sort_order = pymongo.DESCENDING

    return limit, skip, sort_key, sort_order, use_typing


def get_post_data_to_kwargs(request: object) -> dict:
    request_data = request.form.to_dict()

    if not request_data:
        request_data = request.json

    kwargs = {}
    for key, value in request_data.items():
        if "," in value:
            kwargs[key] = {"$in": value.split(",")}
        else:
            kwargs[key] = value

    return kwargs



from fedn.common.config import get_modelstorage_config, get_network_config, get_statestore_config
from fedn.network.controller.control import Control
from fedn.network.storage.statestore.mongostatestore import MongoStateStore

statestore_config = get_statestore_config()
modelstorage_config = get_modelstorage_config()
network_id = get_network_config()

statestore = MongoStateStore(network_id, statestore_config["mongo_config"])
statestore.set_storage_backend(modelstorage_config)
control = Control(statestore=statestore)



import os
import struct
import tempfile

import numpy as np

from fedn.common.log_config import logger
from fedn.utils.helpers.helperbase import HelperBase


class Helper(HelperBase):
    """FEDn helper class for android json model weights."""

    def __init__(self):
        """Initialize helper."""
        self.name = "androidhelper"
        super().__init__()

    # function to calculate an incremental weighted average of the weights

    def increment_average(self, model, model_next, num_examples, total_examples):
        """Incremental weighted average of model weights.

        :param model: Current model weights.
        :type model: list of numpy arrays.
        :param model_next: New model weights.
        :type model_next: list of numpy arrays.
        :param num_examples: Number of examples in new model.
        :type num_examples: int
        :param total_examples: Total number of examples.
        :type total_examples: int
        :return: Incremental weighted average of model weights.
        :rtype: list of numpy arrays.
        """
        # Incremental weighted average
        w = num_examples / total_examples

        return (1 - w) * model + w * model_next

    # function to calculate an incremental weighted average of the weights using numpy.add
    def increment_average_add(self, model, model_next, num_examples, total_examples):
        """Incremental weighted average of model weights.

        :param model: Current model weights.
        :type model: list of numpy arrays.
        :param model_next: New model weights.
        :type model_next: list of numpy arrays.
        :param num_examples: Number of examples in new model.
        :type num_examples: int
        :param total_examples: Total number of examples.
        :type total_examples: int
        :return: Incremental weighted average of model weights.
        :rtype: list of numpy arrays.
        """
        # Incremental weighted average
        w = np.add(
            model,
            num_examples * (np.array(model_next) - np.array(model)) / total_examples,
        )
        return w

    def save(self, weights, path=None):
        """Serialize weights to file. The serialized model must be a single binary object.

        :param weights: weights in json format.
        :param path: Path to file.
        :return: Path to file.
        """
        if not path:
            path = self.get_tmp_path()

        byte_array = struct.pack("f" * len(weights), *weights)
        with open(path, "wb") as file:
            file.write(byte_array)

        return path

    def load(self, fh):
        """Load weights from file or filelike.

        :param fh: file path, filehandle, filelike.
        :return: List of weights in json format.
        """
        logger.debug("in android helper load")
        if isinstance(fh, str):
            with open(fh, "rb") as file:
                byte_data = file.read()
        else:
            byte_data = fh.read()

        weights = np.array(struct.unpack(f"{len(byte_data) // 4}f", byte_data))

        return weights

    def get_tmp_path(self):
        """Return a temporary output path compatible with save_model, load_model.

        :return: Path to file.
        """
        fd, path = tempfile.mkstemp(suffix="")
        os.close(fd)
        return path



import datetime
import uuid
from typing import Any, Dict, List, Tuple

import pymongo
from bson import ObjectId
from pymongo.database import Database

from fedn.network.storage.statestore.stores.store import Store

from .shared import EntityNotFound, from_document


class Session:
    def __init__(self, id: str, session_id: str, status: str, session_config: dict = None):
        self.id = id
        self.session_id = session_id
        self.status = status
        self.session_config = session_config

    def from_dict(data: dict) -> "Session":
        return Session(
            id=str(data["_id"]),
            session_id=data["session_id"] if "session_id" in data else None,
            status=data["status"] if "status" in data else None,
            session_config=data["session_config"] if "session_config" in data else None
        )


class SessionStore(Store[Session]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)

    def _validate_session_config(self, session_config: dict) -> Tuple[bool, str]:
        if "aggregator" not in session_config:
            return False, "session_config.aggregator is required"

        if "round_timeout" not in session_config:
            return False, "session_config.round_timeout is required"

        if not isinstance(session_config["round_timeout"], int):
            return False, "session_config.round_timeout must be an integer"

        if "buffer_size" not in session_config:
            return False, "session_config.buffer_size is required"

        if not isinstance(session_config["buffer_size"], int):
            return False, "session_config.buffer_size must be an integer"

        if "model_id" not in session_config or session_config["model_id"] == "":
            return False, "session_config.model_id is required"

        if not isinstance(session_config["model_id"], str):
            return False, "session_config.model_id must be a string"

        if "delete_models_storage" not in session_config:
            return False, "session_config.delete_models_storage is required"

        if not isinstance(session_config["delete_models_storage"], bool):
            return False, "session_config.delete_models_storage must be a boolean"

        if "clients_required" not in session_config:
            return False, "session_config.clients_required is required"

        if not isinstance(session_config["clients_required"], int):
            return False, "session_config.clients_required must be an integer"

        if "validate" not in session_config:
            return False, "session_config.validate is required"

        if not isinstance(session_config["validate"], bool):
            return False, "session_config.validate must be a boolean"

        if "helper_type" not in session_config:
            return False, "session_config.helper_type is required"

        if not isinstance(session_config["helper_type"], str):
            return False, "session_config.helper_type must be a string"

        return True, ""

    def _validate(self, item: Session) -> Tuple[bool, str]:
        if "session_config" not in item or item["session_config"] is None:
            return False, "session_config is required"
        elif not isinstance(item["session_config"], dict):
            return False, "session_config must be a dict"

        session_config = item["session_config"]

        return  self._validate_session_config(session_config)

    def _complement(self, item: Session):
        item["status"] = "Created"
        item["committed_at"] = datetime.datetime.now()

        if "session_id" not in item or item["session_id"] == "" or not isinstance(item["session_id"], str):
            item["session_id"] = str(uuid.uuid4())


    def get(self, id: str, use_typing: bool = False) -> Session:
        """Get an entity by id
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the session_id (property)
        param use_typing: Whether to return the entity as a typed object or as a dict
            type: bool
        return: The entity
        """
        if ObjectId.is_valid(id):
            id_obj = ObjectId(id)
            document = self.database[self.collection].find_one({"_id": id_obj})
        else:
            document = self.database[self.collection].find_one({"session_id": id})

        if document is None:
            raise EntityNotFound(f"Entity with (id | session_id) {id} not found")

        return Session.from_dict(document) if use_typing else from_document(document)

    def update(self, id: str, item: Session) -> bool:
        raise NotImplementedError("Update not implemented for SessionStore")

    def add(self, item: Session)-> Tuple[bool, Any]:
        """Add an entity
        param item: The entity to add
            type: Session
            description: The entity to add
        return: A tuple with a boolean indicating success and the entity
        """
        valid, message = self._validate(item)
        if not valid:
            return False, message

        self._complement(item)

        return super().add(item)

    def delete(self, id: str) -> bool:
        raise NotImplementedError("Delete not implemented for SessionStore")

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, use_typing: bool = False, **kwargs) -> Dict[int, List[Session]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
            description: The maximum number of entities to return
        param skip: The number of entities to skip
            type: int
            description: The number of entities to skip
        param sort_key: The key to sort by
            type: str
            description: The key to sort by
        param sort_order: The order to sort by
            type: pymongo.DESCENDING
            description: The order to sort by
        param use_typing: Whether to return the entity as a typed object or as a dict
            type: bool
            description: Whether to return the entities as typed objects or as dicts.
        param kwargs: Additional query parameters
            type: dict
            description: Additional query parameters
        return: The entities
        """
        response = super().list(limit, skip, sort_key or "session_id", sort_order, use_typing=use_typing, **kwargs)

        result = [Session.from_dict(item) for item in response["result"]] if use_typing else response["result"]

        return {
            "count": response["count"],
            "result": result
        }



import os
import unittest

import numpy as np

from fedn.utils.helpers.helpers.plugins.numpyhelper import \
    Helper as NumpyHelper


class TestNumpyHelper(unittest.TestCase):
    """Test the PyTorchHelper class."""

    def setUp(self):
        self.helper = NumpyHelper()

    def test_increment_average(self):
        """Test the increment_average method. The weights are stored as list of numpyarray."""

        # Model as OrderedDict with keys as torch layers and values as numpy arrays
        model = [np.array([1, 2, 3])]
        model_next = [np.array([4, 5, 6])]
        a = 10
        W = 20

        result = self.helper.increment_average(model, model_next, a, W)
        print(result)

        # Check  values match
        np.testing.assert_array_equal(result, [np.array([2.5, 3.5, 4.5])])

        # Model as  with keys as torch layers and values as lists
        model = [[1, 2, 3]]
        model_next = [[4, 5, 6]]
        a = 10
        W = 20

        # Catch TypeError: unsupported operand type(s) for -: 'list' and 'list'
        with self.assertRaises(TypeError):
            result = self.helper.increment_average(model, model_next, a, W)

    # Test save and load methods
    def test_save_load(self):
        """Test the save and load methods."""

        # Create a model
        model = [np.array([1, 2, 3])]

        # Save the model
        self.helper.save(model, 'test_model')

        # Check if the model file exists
        self.assertTrue(os.path.exists('test_model.npz'))

        # Load the model
        result = self.helper.load('test_model.npz')

        # Check OrderedDict values match
        np.testing.assert_array_equal(result, [np.array([1, 2, 3])])

        # Remove the model file
        os.remove('test_model.npz')


if __name__ == '__main__':
    unittest.main()



import base64

from fedn.common.log_config import logger
from fedn.network.combiner.interfaces import CombinerInterface
from fedn.network.loadbalancer.leastpacked import LeastPacked

__all__ = ("Network",)


class Network:
    """FEDn network interface. This class is used to interact with the network.
    Note: This class contain redundant code, which is not used in the current version of FEDn.
    Some methods has been moved to :class:`fedn.network.api.interface.API`.
    """

    def __init__(self, control, statestore, load_balancer=None):
        """ """
        self.statestore = statestore
        self.control = control
        self.id = statestore.network_id

        if not load_balancer:
            self.load_balancer = LeastPacked(self)
        else:
            self.load_balancer = load_balancer

    def get_combiner(self, name):
        """Get combiner by name.

        :param name: name of combiner
        :type name: str
        :return: The combiner instance object
        :rtype: :class:`fedn.network.combiner.interfaces.CombinerInterface`
        """
        combiners = self.get_combiners()
        for combiner in combiners:
            if name == combiner.name:
                return combiner
        return None

    def get_combiners(self):
        """Get all combiners in the network.

        :return: list of combiners objects
        :rtype: list(:class:`fedn.network.combiner.interfaces.CombinerInterface`)
        """
        data = self.statestore.get_combiners()
        combiners = []
        for c in data["result"]:
            if c["certificate"]:
                cert = base64.b64decode(c["certificate"])
                key = base64.b64decode(c["key"])
            else:
                cert = None
                key = None

            combiners.append(CombinerInterface(c["parent"], c["name"], c["address"], c["fqdn"], c["port"], certificate=cert, key=key, ip=c["ip"]))

        return combiners

    def add_combiner(self, combiner):
        """Add a new combiner to the network.

        :param combiner: The combiner instance object
        :type combiner: :class:`fedn.network.combiner.interfaces.CombinerInterface`
        :return: None
        """
        if not self.control.idle():
            logger.warning("Reducer is not idle, cannot add additional combiner.")
            return

        if self.get_combiner(combiner.name):
            return

        logger.info("adding combiner {}".format(combiner.name))
        self.statestore.set_combiner(combiner.to_dict())

    def remove_combiner(self, combiner):
        """Remove a combiner from the network.

        :param combiner: The combiner instance object
        :type combiner: :class:`fedn.network.combiner.interfaces.CombinerInterface`
        :return: None
        """
        if not self.control.idle():
            logger.warning("Reducer is not idle, cannot remove combiner.")
            return
        self.statestore.delete_combiner(combiner.name)

    def find_available_combiner(self):
        """Find an available combiner in the network.

        :return: The combiner instance object
        :rtype: :class:`fedn.network.combiner.interfaces.CombinerInterface`
        """
        combiner = self.load_balancer.find_combiner()
        return combiner

    def handle_unavailable_combiner(self, combiner):
        """This callback is triggered if a combiner is found to be unresponsive.

        :param combiner: The combiner instance object
        :type combiner: :class:`fedn.network.combiner.interfaces.CombinerInterface`
        :return: None
        """
        # TODO: Implement strategy to handle an unavailable combiner.
        logger.warning("REDUCER CONTROL: Combiner {} unavailable.".format(combiner.name))

    def add_client(self, client):
        """Add a new client to the network.

        :param client: The client instance object
        :type client: dict
        :return: None
        """
        if self.get_client(client["name"]):
            return

        logger.info("adding client {}".format(client["name"]))
        self.statestore.set_client(client)

    def get_client(self, name):
        """Get client by name.

        :param name: name of client
        :type name: str
        :return: The client instance object
        :rtype: ObjectId
        """
        ret = self.statestore.get_client(name)
        return ret

    def update_client_data(self, client_data, status, role):
        """Update client status in statestore.

        :param client_data: The client instance object
        :type client_data: dict
        :param status: The client status
        :type status: str
        :param role: The client role
        :type role: str
        :return: None
        """
        self.statestore.update_client_status(client_data, status, role)

    def get_client_info(self):
        """List available client in statestore.

        :return: list of client objects
        :rtype: list(ObjectId)
        """
        return self.statestore.list_clients()



from datetime import datetime
from typing import Any, Dict, List, Tuple

import pymongo
from pymongo.database import Database

from fedn.network.storage.statestore.stores.store import Store

from .shared import EntityNotFound, from_document


class Package:
    def __init__(
            self,
            id: str,
            key: str,
            committed_at: datetime,
            description: str,
            file_name: str,
            helper: str,
            name: str,
            storage_file_name: str,
            active: bool = False
    ):
        self.key = key
        self.committed_at = committed_at
        self.description = description
        self.file_name = file_name
        self.helper = helper
        self.id = id
        self.name = name
        self.storage_file_name = storage_file_name
        self.active = active

    def from_dict(data: dict, active_package: dict) -> "Package":
        active = False
        if active_package:
            if "id" in active_package and "id" in data:
                active = active_package["id"] == data["id"]

        return Package(
            id=data["id"] if "id" in data else None,
            key=data["key"] if "key" in data else None,
            committed_at=data["committed_at"] if "committed_at" in data else None,
            description=data["description"] if "description" in data else None,
            file_name=data["file_name"] if "file_name" in data else None,
            helper=data["helper"] if "helper" in data else None,
            name=data["name"] if "name" in data else None,
            storage_file_name=data["storage_file_name"] if "storage_file_name" in data else None,
            active=active
        )


class PackageStore(Store[Package]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)

    def get(self, id: str, use_typing: bool = False) -> Package:
        """Get an entity by id
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the model (property)
        param use_typing: Whether to return the entity as a typed object or as a dict
            type: bool
            description: Whether to return the entities as typed objects or as dicts.
            If True, and active property will be set based on the active package.
        return: The entity
        """
        document = self.database[self.collection].find_one({"id": id})

        if document is None:
            raise EntityNotFound(f"Entity with id {id} not found")

        if not use_typing:
            return from_document(document)

        response_active = self.database[self.collection].find_one({"key": "active"})

        return Package.from_dict(document, response_active)

    def get_active(self, use_typing: bool = False) -> Package:
        """Get the active entity
        param use_typing: Whether to return the entity as a typed object or as a dict
            type: bool
        return: The entity
        """
        response = self.database[self.collection].find_one({"key": "active"})

        if response is None:
            raise EntityNotFound(f"Entity with id {id} not found")

        return Package.from_dict(response, response) if use_typing else from_document(response)

    def update(self, id: str, item: Package) -> bool:
        raise NotImplementedError("Update not implemented for PackageStore")

    def add(self, item: Package)-> Tuple[bool, Any]:
        raise NotImplementedError("Add not implemented for PackageStore")

    def delete(self, id: str) -> bool:
        raise NotImplementedError("Delete not implemented for PackageStore")

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, use_typing: bool = False, **kwargs) -> Dict[int, List[Package]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
        param skip: The number of entities to skip
            type: int
        param sort_key: The key to sort by
            type: str
        param sort_order: The order to sort by
            type: pymongo.DESCENDING | pymongo.ASCENDING
        param use_typing: Whether to return the entities as typed objects or as dicts
            type: bool
            description: Whether to return the entities as typed objects or as dicts.
            If True, and active property will be set based on the active package.
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: A dictionary with the count and the result
        """
        kwargs["key"] = "package_trail"

        response = super().list(limit, skip, sort_key or "committed_at", sort_order, use_typing=True, **kwargs)

        response_active = self.database[self.collection].find_one({"key": "active"})

        result = [Package.from_dict(item, response_active) for item in response["result"]]

        return {
            "count": response["count"],
            "result": result
        }

    def count(self, **kwargs) -> int:
        kwargs["key"] = "package_trail"
        return super().count(**kwargs)



"""Portions of this code are derived from the Apache 2.0 licensed project mlflow (https://mlflow.org/).,
with modifications made by Scaleout Systems AB.
Copyright (c) 2018 Databricks, Inc.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import subprocess
import sys

from fedn.common.log_config import logger

_IS_UNIX = os.name != "nt"


class ShellCommandException(Exception):
    @classmethod
    def from_completed_process(cls, process):
        lines = [
            f"Non-zero exit code: {process.returncode}",
            f"Command: {process.args}",
        ]
        if process.stdout:
            lines += [
                "",
                "STDOUT:",
                process.stdout,
            ]
        if process.stderr:
            lines += [
                "",
                "STDERR:",
                process.stderr,
            ]
        return cls("\n".join(lines))


def _join_commands(*commands):
    entry_point = ["bash", "-c"] if _IS_UNIX else ["cmd", "/c"]
    sep = " && " if _IS_UNIX else " & "
    return [*entry_point, sep.join(map(str, commands))]


def _exec_cmd(
    cmd,
    *,
    cwd=None,
    throw_on_error=True,
    extra_env=None,
    capture_output=True,
    synchronous=True,
    stream_output=False,
    **kwargs,
):
    """A convenience wrapper of `subprocess.Popen` for running a command from a Python script.

    Args:
    ----
        cmd: The command to run, as a string or a list of strings.
        cwd: The current working directory.
        throw_on_error: If True, raises an Exception if the exit code of the program is nonzero.
        extra_env: Extra environment variables to be defined when running the child process.
            If this argument is specified, `kwargs` cannot contain `env`.
        capture_output: If True, stdout and stderr will be captured and included in an exception
            message on failure; if False, these streams won't be captured.
        synchronous: If True, wait for the command to complete and return a CompletedProcess
            instance, If False, does not wait for the command to complete and return
            a Popen instance, and ignore the `throw_on_error` argument.
        stream_output: If True, stream the command's stdout and stderr to `sys.stdout`
            as a unified stream during execution.
            If False, do not stream the command's stdout and stderr to `sys.stdout`.
        kwargs: Keyword arguments (except `text`) passed to `subprocess.Popen`.

    Returns:
    -------
        If synchronous is True, return a `subprocess.CompletedProcess` instance,
        otherwise return a Popen instance.

    """
    illegal_kwargs = set(kwargs.keys()).intersection({"text"})
    if illegal_kwargs:
        raise ValueError(f"`kwargs` cannot contain {list(illegal_kwargs)}")

    env = kwargs.pop("env", None)
    if extra_env is not None and env is not None:
        raise ValueError("`extra_env` and `env` cannot be used at the same time")

    if capture_output and stream_output:
        raise ValueError("`capture_output=True` and `stream_output=True` cannot be specified at the same time")

    env = env if extra_env is None else {**os.environ, **extra_env}

    # In Python < 3.8, `subprocess.Popen` doesn't accept a command containing path-like
    # objects (e.g. `["ls", pathlib.Path("abc")]`) on Windows. To avoid this issue,
    # stringify all elements in `cmd`. Note `str(pathlib.Path("abc"))` returns 'abc'.
    if isinstance(cmd, list):
        cmd = list(map(str, cmd))

    if capture_output or stream_output:
        if kwargs.get("stdout") is not None or kwargs.get("stderr") is not None:
            raise ValueError("stdout and stderr arguments may not be used with capture_output or stream_output")
        kwargs["stdout"] = subprocess.PIPE
        if capture_output:
            kwargs["stderr"] = subprocess.PIPE
        elif stream_output:
            # Redirect stderr to stdout in order to combine the streams for unified printing to
            # `sys.stdout`, as documented in
            # https://docs.python.org/3/library/subprocess.html#subprocess.run
            kwargs["stderr"] = subprocess.STDOUT

    process = subprocess.Popen(
        cmd,
        cwd=cwd,
        env=env,
        text=True,
        **kwargs,
    )
    if not synchronous:
        return process

    if stream_output:
        for output_char in iter(lambda: process.stdout.read(1), ""):
            sys.stdout.write(output_char)

    stdout, stderr = process.communicate()
    returncode = process.poll()
    comp_process = subprocess.CompletedProcess(
        process.args,
        returncode=returncode,
        stdout=stdout,
        stderr=stderr,
    )
    if throw_on_error and returncode != 0:
        raise ShellCommandException.from_completed_process(comp_process)
    return comp_process


def run_process(args, cwd):
    """Run a process and log the output.

    :param args: The arguments to the process.
    :type args: list
    :param cwd: The current working directory.
    :type cwd: str
    :return:
    """
    status = subprocess.Popen(args, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)

    def check_io():
        """Check stdout/stderr of the child process.

        :return:
        """
        while True:
            output = status.stdout.readline().decode()
            if output:
                logger.info(output)
            else:
                break

    # keep checking stdout/stderr until the child exits
    while status.poll() is None:
        check_io()



from typing import Any, Dict, List, Tuple

import pymongo
from pymongo.database import Database

from fedn.network.storage.statestore.stores.store import Store


class Validation:
    def __init__(
            self,
            id: str,
            model_id: str,
            data: str,
            correlation_id: str,
            timestamp: str,
            session_id: str,
            meta: str,
            sender: dict = None,
            receiver: dict = None
    ):
        self.id = id
        self.model_id = model_id
        self.data = data
        self.correlation_id = correlation_id
        self.timestamp = timestamp
        self.session_id = session_id
        self.meta = meta
        self.sender = sender
        self.receiver = receiver

    def from_dict(data: dict) -> "Validation":
        return Validation(
            id=str(data["_id"]),
            model_id=data["modelId"] if "modelId" in data else None,
            data=data["data"] if "data" in data else None,
            correlation_id=data["correlationId"] if "correlationId" in data else None,
            timestamp=data["timestamp"] if "timestamp" in data else None,
            session_id=data["sessionId"] if "sessionId" in data else None,
            meta=data["meta"] if "meta" in data else None,
            sender=data["sender"] if "sender" in data else None,
            receiver=data["receiver"] if "receiver" in data else None
        )


class ValidationStore(Store[Validation]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)

    def get(self, id: str, use_typing: bool = False) -> Validation:
        """Get an entity by id
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the validation (property)
        param use_typing: Whether to return the entity as a typed object or as a dict
            type: bool
        return: The entity
        """
        response = super().get(id, use_typing=use_typing)
        return Validation.from_dict(response) if use_typing else response

    def update(self, id: str, item: Validation) -> bool:
        raise NotImplementedError("Update not implemented for ValidationStore")

    def add(self, item: Validation)-> Tuple[bool, Any]:
        raise NotImplementedError("Add not implemented for ValidationStore")

    def delete(self, id: str) -> bool:
        raise NotImplementedError("Delete not implemented for ValidationStore")

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, use_typing: bool = False, **kwargs) -> Dict[int, List[Validation]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
            description: The maximum number of entities to return
        param skip: The number of entities to skip
            type: int
            description: The number of entities to skip
        param sort_key: The key to sort by
            type: str
            description: The key to sort by
        param sort_order: The order to sort by
            type: pymongo.DESCENDING
            description: The order to sort by
        param use_typing: Whether to return the entities as typed objects or as dicts
            type: bool
            description: Whether to return the entities as typed objects or as dicts
        return: A dictionary with the count and a list of entities
        """
        response = super().list(limit, skip, sort_key or "timestamp", sort_order, use_typing=use_typing, **kwargs)

        result = [Validation.from_dict(item) for item in response["result"]] if use_typing else response["result"]
        return {
            "count": response["count"],
            "result": result
        }



from datetime import datetime
from typing import Any, Dict, List, Tuple

import pymongo
from bson import ObjectId
from pymongo.database import Database

from fedn.network.storage.statestore.stores.store import Store

from .shared import EntityNotFound, from_document


class Model:
    def __init__(self, id: str, key: str, model: str, parent_model: str, session_id: str, committed_at: datetime):
        self.id = id
        self.key = key
        self.model = model
        self.parent_model = parent_model
        self.session_id = session_id
        self.committed_at = committed_at

    def from_dict(data: dict) -> "Model":
        return Model(
            id=str(data["_id"]),
            key=data["key"] if "key" in data else None,
            model=data["model"] if "model" in data else None,
            parent_model=data["parent_model"] if "parent_model" in data else None,
            session_id=data["session_id"] if "session_id" in data else None,
            committed_at=data["committed_at"] if "committed_at" in data else None
        )


class ModelStore(Store[Model]):
    def __init__(self, database: Database, collection: str):
        super().__init__(database, collection)

    def get(self, id: str, use_typing: bool = False) -> Model:
        """Get an entity by id
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the model (property)
        param use_typing: Whether to return the entity as a typed object or as a dict
            type: bool
        return: The entity
        """
        kwargs = {"key": "models"}
        if ObjectId.is_valid(id):
            id_obj = ObjectId(id)
            kwargs["_id"] = id_obj
        else:
            kwargs["model"] = id

        document = self.database[self.collection].find_one(kwargs)

        if document is None:
            raise EntityNotFound(f"Entity with (id | model) {id} not found")

        return Model.from_dict(document) if use_typing else from_document(document)

    def update(self, id: str, item: Model) -> bool:
        raise NotImplementedError("Update not implemented for ModelStore")

    def add(self, item: Model)-> Tuple[bool, Any]:
        raise NotImplementedError("Add not implemented for ModelStore")

    def delete(self, id: str) -> bool:
        raise NotImplementedError("Delete not implemented for ModelStore")

    def list(self, limit: int, skip: int, sort_key: str, sort_order=pymongo.DESCENDING, use_typing: bool = False, **kwargs) -> Dict[int, List[Model]]:
        """List entities
        param limit: The maximum number of entities to return
            type: int
        param skip: The number of entities to skip
            type: int
        param sort_key: The key to sort by
            type: str
        param sort_order: The order to sort by
            type: pymongo.DESCENDING | pymongo.ASCENDING
        param use_typing: Whether to return the entities as typed objects or as dicts
            type: bool
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: A dictionary with the count and the result
        """
        kwargs["key"] = "models"

        response = super().list(limit, skip, sort_key or "committed_at", sort_order, use_typing=use_typing, **kwargs)

        result = [Model.from_dict(item) for item in response["result"]] if use_typing else response["result"]
        return {
            "count": response["count"],
            "result": result
        }

    def list_descendants(self, id: str, limit: int, use_typing: bool = False) -> List[Model]:
        """List descendants
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the model (property)
        param limit: The maximum number of entities to return
            type: int
        param use_typing: Whether to return the entities as typed objects or as dicts
            type: bool
        return: A list of entities
        """
        kwargs = {"key": "models"}
        if ObjectId.is_valid(id):
            id_obj = ObjectId(id)
            kwargs["_id"] = id_obj
        else:
            kwargs["model"] = id

        model: object = self.database[self.collection].find_one(kwargs)

        if model is None:
            raise EntityNotFound(f"Entity with (id | model) {id} not found")

        current_model_id: str = model["model"]
        result: list = []

        for _ in range(limit):
            if current_model_id is None:
                break

            model: str = self.database[self.collection].find_one({"key": "models", "parent_model": current_model_id})

            if model is not None:
                formatted_model = Model.from_dict(model) if use_typing else from_document(model)
                result.append(formatted_model)
                current_model_id = model["model"]
            else:
                break

        result.reverse()

        return result

    def list_ancestors(self, id: str, limit: int, include_self: bool = False, reverse: bool = False, use_typing: bool = False) -> List[Model]:
        """List ancestors
        param id: The id of the entity
            type: str
            description: The id of the entity, can be either the id or the model (property)
        param limit: The maximum number of entities to return
            type: int
        param use_typing: Whether to return the entities as typed objects or as dicts
            type: bool
        return: A list of entities
        """
        kwargs = {"key": "models"}
        if ObjectId.is_valid(id):
            id_obj = ObjectId(id)
            kwargs["_id"] = id_obj
        else:
            kwargs["model"] = id

        model: object = self.database[self.collection].find_one(kwargs)

        if model is None:
            raise EntityNotFound(f"Entity with (id | model) {id} not found")

        current_model_id: str = model["parent_model"]
        result: list = []

        if include_self:
            formatted_model = Model.from_dict(model) if use_typing else from_document(model)
            result.append(formatted_model)

        for _ in range(limit):
            if current_model_id is None:
                break

            model = self.database[self.collection].find_one({"key": "models", "model": current_model_id})

            if model is not None:
                formatted_model = Model.from_dict(model) if use_typing else from_document(model)
                result.append(formatted_model)
                current_model_id = model["parent_model"]
            else:
                break

        if reverse:
            result.reverse()

        return result

    def count(self, **kwargs) -> int:
        """Count entities
        param kwargs: Additional query parameters
            type: dict
            example: {"key": "models"}
        return: The count (int)
        """
        kwargs["key"] = "models"
        return super().count(**kwargs)



import base64
import io
import json
import os
import queue
import re
import socket
import sys
import threading
import time
import uuid
from datetime import datetime
from io import BytesIO
from shutil import copytree

import grpc
from cryptography.hazmat.primitives.serialization import Encoding
from google.protobuf.json_format import MessageToJson
from OpenSSL import SSL
from tenacity import retry, stop_after_attempt

import fedn.network.grpc.fedn_pb2 as fedn
import fedn.network.grpc.fedn_pb2_grpc as rpc
from fedn.common.config import FEDN_AUTH_SCHEME, FEDN_PACKAGE_EXTRACT_DIR
from fedn.common.log_config import (logger, set_log_level_from_string,
                                    set_log_stream)
from fedn.network.clients.connect import ConnectorClient, Status
from fedn.network.clients.package import PackageRuntime
from fedn.network.clients.state import ClientState, ClientStateToString
from fedn.network.combiner.modelservice import (get_tmp_path,
                                                upload_request_generator)
from fedn.utils.dispatcher import Dispatcher
from fedn.utils.helpers.helpers import get_helper

CHUNK_SIZE = 1024 * 1024
VALID_NAME_REGEX = "^[a-zA-Z0-9_-]*$"


class GrpcAuth(grpc.AuthMetadataPlugin):
    def __init__(self, key):
        self._key = key

    def __call__(self, context, callback):
        callback((("authorization", f"{FEDN_AUTH_SCHEME} {self._key}"),), None)


class Client:
    """FEDn Client. Service running on client/datanodes in a federation,
    recieving and handling model update and model validation requests.

    :param config: A configuration dictionary containing connection information for the discovery service (controller)
        and settings governing e.g. client-combiner assignment behavior.
    :type config: dict
    """

    def __init__(self, config):
        """Initialize the client."""
        self.state = None
        self.error_state = False
        self._connected = False
        self._missed_heartbeat = 0
        self.config = config
        self.trace_attribs = False
        set_log_level_from_string(config.get("verbosity", "INFO"))
        set_log_stream(config.get("logfile", None))

        self.connector = ConnectorClient(
            host=config["discover_host"],
            port=config["discover_port"],
            token=config["token"],
            name=config["name"],
            remote_package=config["remote_compute_context"],
            force_ssl=config["force_ssl"],
            verify=config["verify"],
            combiner=config["preferred_combiner"],
            id=config["client_id"],
        )

        # Validate client name
        match = re.search(VALID_NAME_REGEX, config["name"])
        if not match:
            raise ValueError("Unallowed character in client name. Allowed characters: a-z, A-Z, 0-9, _, -.")

        # Folder where the client will store downloaded compute package and logs
        self.name = config["name"]
        if FEDN_PACKAGE_EXTRACT_DIR:
            self.run_path = os.path.join(os.getcwd(), FEDN_PACKAGE_EXTRACT_DIR)
        else:
            dirname = self.name + "-" + time.strftime("%Y%m%d-%H%M%S")
            self.run_path = os.path.join(os.getcwd(), dirname)
        if not os.path.exists(self.run_path):
            os.mkdir(self.run_path)

        self.started_at = datetime.now()
        self.logs = []

        self.inbox = queue.Queue()

        # Attach to the FEDn network (get combiner)
        combiner_config = self.assign()
        self.connect(combiner_config)

        self._initialize_dispatcher(self.config)

        self._initialize_helper(combiner_config)
        if not self.helper:
            logger.warning("Failed to retrieve helper class settings: {}".format(combiner_config))

        self._subscribe_to_combiner(self.config)

        self.state = ClientState.idle

    def assign(self):
        """Contacts the controller and asks for combiner assignment.

        :return: A configuration dictionary containing connection information for combiner.
        :rtype: dict
        """
        logger.info("Initiating assignment request.")
        while True:
            status, response = self.connector.assign()
            if status == Status.TryAgain:
                logger.warning(response)
                logger.info("Assignment request failed. Retrying in 5 seconds.")
                time.sleep(5)
                continue
            if status == Status.Assigned:
                combiner_config = response
                break
            if status == Status.UnAuthorized:
                logger.critical(response)
                sys.exit("Exiting: Unauthorized")
            if status == Status.UnMatchedConfig:
                logger.critical(response)
                sys.exit("Exiting: UnMatchedConfig")
            time.sleep(5)
        logger.info("Assignment successfully received.")
        logger.info("Received combiner configuration: {}".format(combiner_config))
        return combiner_config

    def _add_grpc_metadata(self, key, value):
        """Add metadata for gRPC calls.

        :param key: The key of the metadata.
        :type key: str
        :param value: The value of the metadata.
        :type value: str
        """
        # Check if metadata exists and add if not
        if not hasattr(self, "metadata"):
            self.metadata = ()

        # Check if metadata key already exists and replace value if so
        for i, (k, v) in enumerate(self.metadata):
            if k == key:
                # Replace value
                self.metadata = self.metadata[:i] + ((key, value),) + self.metadata[i + 1 :]
                return

        # Set metadata using tuple concatenation
        self.metadata += ((key, value),)

    def _get_ssl_certificate(self, domain, port=443):
        context = SSL.Context(SSL.SSLv23_METHOD)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((domain, port))
        ssl_sock = SSL.Connection(context, sock)
        ssl_sock.set_tlsext_host_name(domain.encode())
        ssl_sock.set_connect_state()
        ssl_sock.do_handshake()
        cert = ssl_sock.get_peer_certificate()
        ssl_sock.close()
        sock.close()
        cert = cert.to_cryptography().public_bytes(Encoding.PEM).decode()
        return cert

    def connect(self, combiner_config):
        """Connect to combiner.

        :param combiner_config: connection information for the combiner.
        :type combiner_config: dict
        """
        if self._connected:
            logger.info("Client is already attached. ")
            return

        # TODO use the combiner_config['certificate'] for setting up secure comms'
        host = combiner_config["host"]
        # Add host to gRPC metadata
        self._add_grpc_metadata("grpc-server", host)
        logger.debug("Client using metadata: {}.".format(self.metadata))
        port = combiner_config["port"]
        secure = False
        if combiner_config["fqdn"] is not None:
            host = combiner_config["fqdn"]
            # assuming https if fqdn is used
            port = 443
        logger.info(f"Initiating connection to combiner host at: {host}:{port}")

        if combiner_config["certificate"]:
            logger.info("Utilizing CA certificate for GRPC channel authentication.")
            secure = True
            cert = base64.b64decode(combiner_config["certificate"])  # .decode('utf-8')
            credentials = grpc.ssl_channel_credentials(root_certificates=cert)
            channel = grpc.secure_channel("{}:{}".format(host, str(port)), credentials)
        elif os.getenv("FEDN_GRPC_ROOT_CERT_PATH"):
            secure = True
            logger.info("Using root certificate from environment variable for GRPC channel.")
            with open(os.environ["FEDN_GRPC_ROOT_CERT_PATH"], "rb") as f:
                credentials = grpc.ssl_channel_credentials(f.read())
            channel = grpc.secure_channel("{}:{}".format(host, str(port)), credentials)
        elif self.config["secure"]:
            secure = True
            logger.info("Using CA certificate for GRPC channel.")
            cert = self._get_ssl_certificate(host, port=port)

            credentials = grpc.ssl_channel_credentials(cert.encode("utf-8"))
            if self.config["token"]:
                token = self.config["token"]
                auth_creds = grpc.metadata_call_credentials(GrpcAuth(token))
                channel = grpc.secure_channel("{}:{}".format(host, str(port)), grpc.composite_channel_credentials(credentials, auth_creds))
            else:
                channel = grpc.secure_channel("{}:{}".format(host, str(port)), credentials)
        else:
            logger.info("Using insecure GRPC channel.")
            if port == 443:
                port = 80
            channel = grpc.insecure_channel("{}:{}".format(host, str(port)))

        self.channel = channel

        self.connectorStub = rpc.ConnectorStub(channel)
        self.combinerStub = rpc.CombinerStub(channel)
        self.modelStub = rpc.ModelServiceStub(channel)

        logger.info("Successfully established {} connection to {}:{}".format("secure" if secure else "insecure", host, port))

        logger.info("Using {} compute package.".format(combiner_config["package"]))

        self._connected = True

    def disconnect(self):
        """Disconnect from the combiner."""
        if not self._connected:
            logger.info("Client is not connected.")

        self.channel.close()
        self._connected = False
        logger.info("Client {} disconnected.".format(self.name))

    def _initialize_helper(self, combiner_config):
        """Initialize the helper class for the client.

        :param combiner_config: A configuration dictionary containing connection information for
        | the discovery service (controller) and settings governing e.g.
        | client-combiner assignment behavior.
        :type combiner_config: dict
        :return:
        """
        if "helper_type" in combiner_config.keys():
            self.helper = get_helper(combiner_config["helper_type"])

    def _subscribe_to_combiner(self, config):
        """Listen to combiner message stream and start all processing threads.

        :param config: A configuration dictionary containing connection information for
        | the discovery service (controller) and settings governing e.g.
        | client-combiner assignment behavior.
        """
        # Start sending heartbeats to the combiner.
        threading.Thread(target=self._send_heartbeat, kwargs={"update_frequency": config["heartbeat_interval"]}, daemon=True).start()

        # Start listening for combiner training and validation messages
        threading.Thread(target=self._listen_to_task_stream, daemon=True).start()
        self._connected = True

        # Start processing the client message inbox
        threading.Thread(target=self.process_request, daemon=True).start()

    @retry(stop=stop_after_attempt(3))
    def untar_package(self, package_runtime):
        _, package_runpath = package_runtime.unpack()
        return package_runpath

    def _initialize_dispatcher(self, config):
        """Initialize the dispatcher for the client.

        :param config: A configuration dictionary containing connection information for
        | the discovery service (controller) and settings governing e.g.
        | client-combiner assignment behavior.
        :type config: dict
        :return:
        """
        if config["remote_compute_context"]:
            pr = PackageRuntime(self.run_path)

            retval = None
            tries = 10

            while tries > 0:
                retval = pr.download(
                    host=config["discover_host"], port=config["discover_port"], token=config["token"], force_ssl=config["force_ssl"], secure=config["secure"]
                )
                if retval:
                    break
                time.sleep(60)
                logger.warning("Compute package not available. Retrying in 60 seconds. {} attempts remaining.".format(tries))
                tries -= 1

            if retval:
                if "checksum" not in config:
                    logger.warning("Bypassing validation of package checksum. Ensure the package source is trusted.")
                else:
                    checks_out = pr.validate(config["checksum"])
                    if not checks_out:
                        logger.critical("Validation of local package failed. Client terminating.")
                        self.error_state = True
                        return
            package_runpath = ""
            if retval:
                package_runpath = self.untar_package(pr)

            self.dispatcher = pr.dispatcher(package_runpath)
            try:
                logger.info("Initiating Dispatcher with entrypoint set to: startup")
                activate_cmd = self.dispatcher._get_or_create_python_env()
                self.dispatcher.run_cmd("startup")
            except KeyError:
                logger.info("No startup command found in package. Continuing.")
                pass
            except Exception as e:
                logger.error(f"Caught exception: {type(e).__name__}")

        else:
            # TODO: Deprecate
            dispatch_config = {
                "entry_points": {
                    "predict": {"command": "python3 predict.py"},
                    "train": {"command": "python3 train.py"},
                    "validate": {"command": "python3 validate.py"},
                }
            }
            from_path = os.path.join(os.getcwd(), "client")

            copytree(from_path, self.run_path)
            self.dispatcher = Dispatcher(dispatch_config, self.run_path)
        # Get or create python environment
        activate_cmd = self.dispatcher._get_or_create_python_env()
        if activate_cmd:
            logger.info("To activate the virtual environment, run: {}".format(activate_cmd))

    def get_model_from_combiner(self, id, timeout=20):
        """Fetch a model from the assigned combiner.
        Downloads the model update object via a gRPC streaming channel.

        :param id: The id of the model update object.
        :type id: str
        :return: The model update object.
        :rtype: BytesIO
        """
        data = BytesIO()
        time_start = time.time()
        request = fedn.ModelRequest(id=id)
        request.sender.name = self.name
        request.sender.role = fedn.WORKER

        try:
            for part in self.modelStub.Download(request, metadata=self.metadata):
                if part.status == fedn.ModelStatus.IN_PROGRESS:
                    data.write(part.data)

                if part.status == fedn.ModelStatus.OK:
                    return data

                if part.status == fedn.ModelStatus.FAILED:
                    return None

                if part.status == fedn.ModelStatus.UNKNOWN:
                    if time.time() - time_start >= timeout:
                        return None
                    continue
        except grpc.RpcError as e:
            logger.critical(f"GRPC: An error occurred during model download: {e}")

        return data

    def send_model_to_combiner(self, model, id):
        """Send a model update to the assigned combiner.
        Uploads the model updated object via a gRPC streaming channel, Upload.

        :param model: The model update object.
        :type model: BytesIO
        :param id: The id of the model update object.
        :type id: str
        :return: The model update object.
        :rtype: BytesIO
        """
        if not isinstance(model, BytesIO):
            bt = BytesIO()

            for d in model.stream(32 * 1024):
                bt.write(d)
        else:
            bt = model

        bt.seek(0, 0)

        try:
            result = self.modelStub.Upload(upload_request_generator(bt, id), metadata=self.metadata)
        except grpc.RpcError as e:
            logger.critical(f"GRPC: An error occurred during model upload: {e}")

        return result

    def _listen_to_task_stream(self):
        """Subscribe to the model update request stream.

        :return: None
        :rtype: None
        """
        r = fedn.ClientAvailableMessage()
        r.sender.name = self.name
        r.sender.role = fedn.WORKER
        # Add client to metadata
        self._add_grpc_metadata("client", self.name)

        while self._connected:
            try:
                for request in self.combinerStub.TaskStream(r, metadata=self.metadata):
                    if request:
                        logger.debug("Received model update request from combiner: {}.".format(request))
                    if request.sender.role == fedn.COMBINER:
                        # Process training request
                        self.send_status(
                            "Received model update request.",
                            log_level=fedn.Status.AUDIT,
                            type=fedn.StatusType.MODEL_UPDATE_REQUEST,
                            request=request,
                            sesssion_id=request.session_id,
                        )
                        logger.info("Received model update request of type {} for model_id {}".format(request.type, request.model_id))

                        if request.type == fedn.StatusType.MODEL_UPDATE and self.config["trainer"]:
                            self.inbox.put(("train", request))
                        elif request.type == fedn.StatusType.MODEL_VALIDATION and self.config["validator"]:
                            self.inbox.put(("validate", request))
                        else:
                            logger.error("Unknown request type: {}".format(request.type))

            except grpc.RpcError as e:
                # Handle gRPC errors
                status_code = e.code()
                if status_code == grpc.StatusCode.UNAVAILABLE:
                    logger.warning("GRPC TaskStream: server unavailable during model update request stream. Retrying.")
                    # Retry after a delay
                    time.sleep(5)
                if status_code == grpc.StatusCode.UNAUTHENTICATED:
                    details = e.details()
                    if details == "Token expired":
                        logger.warning("GRPC TaskStream: Token expired. Reconnecting.")
                        self.disconnect()

                if status_code == grpc.StatusCode.CANCELLED:
                    # Expected if the client is disconnected
                    logger.critical("GRPC TaskStream: Client disconnected from combiner. Trying to reconnect.")

                else:
                    # Log the error and continue
                    logger.error(f"GRPC TaskStream: An error occurred during model update request stream: {e}")

            except Exception as ex:
                # Handle other exceptions
                logger.error(f"GRPC TaskStream: An error occurred during model update request stream: {ex}")

        # Detach if not attached
        if not self._connected:
            return

    def _process_training_request(self, model_id: str, session_id: str = None):
        """Process a training (model update) request.

        :param model_id: The model id of the model to be updated.
        :type model_id: str
        :param session_id: The id of the current session
        :type session_id: str
        :return: The model id of the updated model, or None if the update failed. And a dict with metadata.
        :rtype: tuple
        """
        self.send_status("\t Starting processing of training request for model_id {}".format(model_id), sesssion_id=session_id)
        self.state = ClientState.training

        try:
            meta = {}
            tic = time.time()
            mdl = self.get_model_from_combiner(str(model_id))
            if mdl is None:
                logger.error("Could not retrieve model from combiner. Aborting training request.")
                return None, None
            meta["fetch_model"] = time.time() - tic

            inpath = self.helper.get_tmp_path()
            with open(inpath, "wb") as fh:
                fh.write(mdl.getbuffer())

            outpath = self.helper.get_tmp_path()
            tic = time.time()
            # TODO: Check return status, fail gracefully

            self.dispatcher.run_cmd("train {} {}".format(inpath, outpath))

            meta["exec_training"] = time.time() - tic

            tic = time.time()
            out_model = None

            with open(outpath, "rb") as fr:
                out_model = io.BytesIO(fr.read())

            # Stream model update to combiner server
            updated_model_id = uuid.uuid4()
            self.send_model_to_combiner(out_model, str(updated_model_id))
            meta["upload_model"] = time.time() - tic

            # Read the metadata file
            with open(outpath + "-metadata", "r") as fh:
                training_metadata = json.loads(fh.read())
            meta["training_metadata"] = training_metadata

            os.unlink(inpath)
            os.unlink(outpath)
            os.unlink(outpath + "-metadata")

        except Exception as e:
            logger.error("Could not process training request due to error: {}".format(e))
            updated_model_id = None
            meta = {"status": "failed", "error": str(e)}

        self.state = ClientState.idle

        return updated_model_id, meta

    def _process_validation_request(self, model_id: str, is_inference: bool, session_id: str = None):
        """Process a validation request.

        :param model_id: The model id of the model to be validated.
        :type model_id: str
        :param is_inference: True if the validation is an inference request, False if it is a validation request.
        :type is_inference: bool
        :param session_id: The id of the current session.
        :type session_id: str
        :return: The validation metrics, or None if validation failed.
        :rtype: dict
        """
        # Figure out cmd
        if is_inference:
            cmd = "infer"
        else:
            cmd = "validate"

        self.send_status(f"Processing {cmd} request for model_id {model_id}", sesssion_id=session_id)
        self.state = ClientState.validating
        try:
            model = self.get_model_from_combiner(str(model_id))
            if model is None:
                logger.error("Could not retrieve model from combiner. Aborting validation request.")
                return None
            inpath = self.helper.get_tmp_path()

            with open(inpath, "wb") as fh:
                fh.write(model.getbuffer())

            outpath = get_tmp_path()
            self.dispatcher.run_cmd(f"{cmd} {inpath} {outpath}")

            with open(outpath, "r") as fh:
                validation = json.loads(fh.read())

            os.unlink(inpath)
            os.unlink(outpath)

        except Exception as e:
            logger.warning("Validation failed with exception {}".format(e))
            self.state = ClientState.idle
            return None

        self.state = ClientState.idle
        return validation

    def process_request(self):
        """Process training and validation tasks."""
        while True:
            if not self._connected:
                return

            try:
                (task_type, request) = self.inbox.get(timeout=1.0)
                if task_type == "train":
                    tic = time.time()
                    self.state = ClientState.training
                    model_id, meta = self._process_training_request(request.model_id, session_id=request.session_id)

                    if meta is not None:
                        processing_time = time.time() - tic
                        meta["processing_time"] = processing_time
                        meta["config"] = request.data

                    if model_id is not None:
                        # Send model update to combiner
                        update = fedn.ModelUpdate()
                        update.sender.name = self.name
                        update.sender.role = fedn.WORKER
                        update.receiver.name = request.sender.name
                        update.receiver.role = request.sender.role
                        update.model_id = request.model_id
                        update.model_update_id = str(model_id)
                        update.timestamp = str(datetime.now())
                        update.correlation_id = request.correlation_id
                        update.meta = json.dumps(meta)

                        try:
                            _ = self.combinerStub.SendModelUpdate(update, metadata=self.metadata)
                            self.send_status(
                                "Model update completed.",
                                log_level=fedn.Status.AUDIT,
                                type=fedn.StatusType.MODEL_UPDATE,
                                request=update,
                                sesssion_id=request.session_id,
                            )
                        except grpc.RpcError as e:
                            status_code = e.code()
                            logger.error("GRPC error, {}.".format(status_code.name))
                            logger.debug(e)
                        except ValueError as e:
                            logger.error("GRPC error, RPC channel closed. {}".format(e))
                            logger.debug(e)
                    else:
                        self.send_status(
                            "Client {} failed to complete model update.", log_level=fedn.Status.WARNING, request=request, sesssion_id=request.session_id
                        )

                    self.state = ClientState.idle
                    self.inbox.task_done()

                elif task_type == "validate":
                    self.state = ClientState.validating
                    metrics = self._process_validation_request(request.model_id, False, request.session_id)

                    if metrics is not None:
                        # Send validation
                        validation = fedn.ModelValidation()
                        validation.sender.name = self.name
                        validation.sender.role = fedn.WORKER
                        validation.receiver.name = request.sender.name
                        validation.receiver.role = request.sender.role
                        validation.model_id = str(request.model_id)
                        validation.data = json.dumps(metrics)
                        validation.timestamp.GetCurrentTime()
                        validation.correlation_id = request.correlation_id
                        validation.session_id = request.session_id

                        try:
                            _ = self.combinerStub.SendModelValidation(validation, metadata=self.metadata)

                            status_type = fedn.StatusType.MODEL_VALIDATION
                            self.send_status(
                                "Model validation completed.", log_level=fedn.Status.AUDIT, type=status_type, request=validation, sesssion_id=request.session_id
                            )
                        except grpc.RpcError as e:
                            status_code = e.code()
                            logger.error("GRPC error, {}.".format(status_code.name))
                            logger.debug(e)
                        except ValueError as e:
                            logger.error("GRPC error, RPC channel closed. {}".format(e))
                            logger.debug(e)
                    else:
                        self.send_status(
                            "Client {} failed to complete model validation.".format(self.name),
                            log_level=fedn.Status.WARNING,
                            request=request,
                            sesssion_id=request.session_id,
                        )

                    self.state = ClientState.idle
                    self.inbox.task_done()
            except queue.Empty:
                pass
            except grpc.RpcError as e:
                logger.critical(f"GRPC process_request: An error occurred during process request: {e}")

    def _send_heartbeat(self, update_frequency=2.0):
        """Send a heartbeat to the combiner.

        :param update_frequency: The frequency of the heartbeat in seconds.
        :type update_frequency: float
        :return: None if the client is disconnected.
        :rtype: None
        """
        while True:
            heartbeat = fedn.Heartbeat(sender=fedn.Client(name=self.name, role=fedn.WORKER))
            try:
                self.connectorStub.SendHeartbeat(heartbeat, metadata=self.metadata)
                self._missed_heartbeat = 0
            except grpc.RpcError as e:
                status_code = e.code()
                if status_code == grpc.StatusCode.UNAVAILABLE:
                    self._missed_heartbeat += 1
                    logger.error(
                        "GRPC hearbeat: combiner unavailable, retrying (attempt {}/{}).".format(
                            self._missed_heartbeat, self.config["reconnect_after_missed_heartbeat"]
                        )
                    )
                    if self._missed_heartbeat > self.config["reconnect_after_missed_heartbeat"]:
                        self.disconnect()
                        self._missed_heartbeat = 0
                if status_code == grpc.StatusCode.UNAUTHENTICATED:
                    details = e.details()
                    if details == "Token expired":
                        logger.error("GRPC hearbeat: Token expired. Disconnecting.")
                        self.disconnect()
                        sys.exit("Unauthorized. Token expired. Please obtain a new token.")
                logger.debug(e)

            time.sleep(update_frequency)
            if not self._connected:
                logger.info("SendStatus: Client disconnected.")
                return

    def send_status(self, msg, log_level=fedn.Status.INFO, type=None, request=None, sesssion_id: str = None):
        """Send status message.

        :param msg: The message to send.
        :type msg: str
        :param log_level: The log level of the message.
        :type log_level: fedn.Status.INFO, fedn.Status.WARNING, fedn.Status.ERROR
        :param type: The type of the message.
        :type type: str
        :param request: The request message.
        :type request: fedn.Request
        """
        if not self._connected:
            logger.info("SendStatus: Client disconnected.")
            return

        status = fedn.Status()
        status.timestamp.GetCurrentTime()
        status.sender.name = self.name
        status.sender.role = fedn.WORKER
        status.log_level = log_level
        status.status = str(msg)
        status.session_id = sesssion_id
        if type is not None:
            status.type = type

        if request is not None:
            status.data = MessageToJson(request)

        self.logs.append("{} {} LOG LEVEL {} MESSAGE {}".format(str(datetime.now()), status.sender.name, status.log_level, status.status))
        try:
            _ = self.connectorStub.SendStatus(status, metadata=self.metadata)
        except grpc.RpcError as e:
            status_code = e.code()
            if status_code == grpc.StatusCode.UNAVAILABLE:
                logger.warning("GRPC SendStatus: server unavailable during send status.")
            if status_code == grpc.StatusCode.UNAUTHENTICATED:
                details = e.details()
                if details == "Token expired":
                    logger.warning("GRPC SendStatus: Token expired.")

    def run(self):
        """Run the client."""
        try:
            cnt = 0
            old_state = self.state
            while True:
                time.sleep(1)
                if cnt == 0:
                    logger.info("Client is active, waiting for model update requests.")
                    cnt = 1
                if self.state != old_state:
                    logger.info("Client in {} state.".format(ClientStateToString(self.state)))
                if not self._connected:
                    logger.warning("Client lost connection to combiner. Attempting to reconnect to FEDn network.")
                    combiner_config = self.assign()
                    self.connect(combiner_config)
                    self._subscribe_to_combiner(self.config)
                    cnt = 0
                if self.error_state:
                    logger.error("Client in error state. Terminiating.")
                    sys.exit("Client in error state. Terminiating.")
        except KeyboardInterrupt:
            logger.info("Shutting down.")



import grpc
import jwt

from fedn.common.config import FEDN_AUTH_SCHEME, FEDN_JWT_ALGORITHM, SECRET_KEY
from fedn.common.log_config import logger
from fedn.network.api.auth import check_custom_claims

ENDPOINT_ROLES_MAPPING = {
    "/fedn.Combiner/TaskStream": ["client"],
    "/fedn.Combiner/SendModelUpdate": ["client"],
    "/fedn.Combiner/SendModelValidation": ["client"],
    "/fedn.Connector/SendHeartbeat": ["client"],
    "/fedn.Connector/SendStatus": ["client"],
    "/fedn.ModelService/Download": ["client"],
    "/fedn.ModelService/Upload": ["client"],
    "/fedn.Control/Start": ["controller"],
    "/fedn.Control/Stop": ["controller"],
    "/fedn.Control/FlushAggregationQueue": ["controller"],
    "/fedn.Control/SetAggregator": ["controller"],
}

ENDPOINT_WHITELIST = [
    "/fedn.Connector/AcceptingClients",
    "/fedn.Connector/ListActiveClients",
    "/fedn.Control/Start",
    "/fedn.Control/Stop",
    "/fedn.Control/FlushAggregationQueue",
    "/fedn.Control/SetAggregator",
]

USER_AGENT_WHITELIST = ["grpc_health_probe"]


def check_role_claims(payload, endpoint):
    user_role = payload.get("role", "")

    # Perform endpoint-specific RBAC check
    allowed_roles = ENDPOINT_ROLES_MAPPING.get(endpoint)
    if allowed_roles and user_role not in allowed_roles:
        return False
    return True


def _unary_unary_rpc_terminator(code, details):
    def terminate(ignored_request, context):
        context.abort(code, details)

    return grpc.unary_unary_rpc_method_handler(terminate)


class JWTInterceptor(grpc.ServerInterceptor):
    def __init__(self):
        pass

    def intercept_service(self, continuation, handler_call_details):
        # Pass if no secret key is set
        if not SECRET_KEY:
            return continuation(handler_call_details)
        metadata = dict(handler_call_details.invocation_metadata)
        # Pass whitelisted methods
        if handler_call_details.method in ENDPOINT_WHITELIST:
            return continuation(handler_call_details)
        # Pass if the request comes from whitelisted user agents
        user_agent = metadata.get("user-agent").split(" ")[0]
        if user_agent in USER_AGENT_WHITELIST:
            return continuation(handler_call_details)

        token = metadata.get("authorization")
        if token is None:
            return _unary_unary_rpc_terminator(grpc.StatusCode.UNAUTHENTICATED, "Token is missing")

        if not token.startswith(FEDN_AUTH_SCHEME):
            return _unary_unary_rpc_terminator(grpc.StatusCode.UNAUTHENTICATED, f"Invalid token scheme, expected {FEDN_AUTH_SCHEME}")

        token = token.split(" ")[1]

        try:
            payload = jwt.decode(token, SECRET_KEY, algorithms=[FEDN_JWT_ALGORITHM])

            if not check_role_claims(payload, handler_call_details.method):
                return _unary_unary_rpc_terminator(grpc.StatusCode.PERMISSION_DENIED, "Insufficient permissions")

            if not check_custom_claims(payload):
                return _unary_unary_rpc_terminator(grpc.StatusCode.PERMISSION_DENIED, "Insufficient permissions")

            return continuation(handler_call_details)
        except jwt.InvalidTokenError:
            return _unary_unary_rpc_terminator(grpc.StatusCode.UNAUTHENTICATED, "Invalid token")
        except jwt.ExpiredSignatureError:
            return _unary_unary_rpc_terminator(grpc.StatusCode.UNAUTHENTICATED, "Token expired")
        except Exception as e:
            logger.error(str(e))
            return _unary_unary_rpc_terminator(grpc.StatusCode.UNKNOWN, str(e))



import io
from collections import defaultdict
from io import BytesIO

from fedn.network.storage.models.modelstorage import ModelStorage

CHUNK_SIZE = 1024 * 1024


class MemoryModelStorage(ModelStorage):
    """Class for in-memory storage of model artifacts.

    Models are stored as BytesIO objects in a dictionary.

    """

    def __init__(self):
        self.models = defaultdict(io.BytesIO)
        self.models_metadata = {}

    def exist(self, model_id):
        if model_id in self.models.keys():
            return True
        return False

    def get(self, model_id):
        obj = self.models[model_id]
        obj.seek(0, 0)
        # Have to copy object to not mix up the file pointers when sending... fix in better way.
        obj = BytesIO(obj.read())
        return obj

    def get_ptr(self, model_id):
        """:param model_id:
        :return:
        """
        return self.models[model_id]

    def get_model_metadata(self, model_id):
        return self.models_metadata[model_id]

    def set_model_metadata(self, model_id, model_metadata):
        self.models_metadata.update({model_id: model_metadata})



import threading

from flask import Blueprint, jsonify, request

from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.v1.shared import api_version, get_post_data_to_kwargs, get_typed_list_headers, mdb
from fedn.network.storage.statestore.stores.session_store import SessionStore
from fedn.network.storage.statestore.stores.shared import EntityNotFound
from .model_routes import model_store
from fedn.network.api.shared import control

bp = Blueprint("session", __name__, url_prefix=f"/api/{api_version}/sessions")

session_store = SessionStore(mdb, "control.sessions")


@bp.route("/", methods=["GET"])
@jwt_auth_required(role="admin")
def get_sessions():
    """Get sessions
    Retrieves a list of sessions based on the provided parameters.
    By specifying a parameter in the url, you can filter the sessions based on that parameter,
    and the response will contain only the sessions that match the filter.
    ---
    tags:
        - Sessions
    parameters:
      - name: status
        in: query
        required: false
        type: string
        description: Status of the session
      - name: session_id
        in: query
        required: false
        type: string
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of sessions to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of sessions to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the sessions by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the sessions in ('asc' or 'desc')
    definitions:
      Session:
        type: object
        properties:
          id:
            type: string
          status:
            type: string
          session_id:
            type: string
          session_config:
            type: object
    responses:
      200:
        description: A list of sessions and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Session'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, _ = get_typed_list_headers(request.headers)
        kwargs = request.args.to_dict()

        sessions = session_store.list(limit, skip, sort_key, sort_order, use_typing=False, **kwargs)

        result = sessions["result"]

        response = {"count": sessions["count"], "result": result}

        return jsonify(response), 200
    except Exception as e:
        return jsonify({"message": str(e)}), 500


@bp.route("/list", methods=["POST"])
@jwt_auth_required(role="admin")
def list_sessions():
    """List sessions
    Retrieves a list of sessions based on the provided parameters.
    Works much like the GET /sessions endpoint, but allows for more complex queries.
    By specifying a parameter in the request body, you can filter the sessions based on that parameter,
    and the response will contain only the sessions that match the filter. If the parameter value contains a comma,
    the filter will be an "in" query, meaning that the sessions will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
      - Sessions
    parameters:
      - name: session
        in: body
        required: false
        schema:
          type: object
          properties:
            status:
              type: string
            session_id:
              type: string
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of sessions to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of sessions to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the sessions by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the sessions in ('asc' or 'desc')
    responses:
      200:
        description: A list of sessions and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Session'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, _ = get_typed_list_headers(request.headers)
        kwargs = get_post_data_to_kwargs(request)

        sessions = session_store.list(limit, skip, sort_key, sort_order, use_typing=False, **kwargs)

        result = sessions["result"]

        response = {"count": sessions["count"], "result": result}

        return jsonify(response), 200
    except Exception as e:
        return jsonify({"message": str(e)}), 500


@bp.route("/count", methods=["GET"])
@jwt_auth_required(role="admin")
def get_sessions_count():
    """Sessions count
    Retrieves the count of sessions based on the provided parameters.
    By specifying a parameter in the url, you can filter the sessions based on that parameter,
    and the response will contain only the count of sessions that match the filter.
    ---
    tags:
        - Sessions
    parameters:
      - name: session_id
        in: query
        required: false
        type: string
      - name: status
        in: query
        required: false
        type: string
    responses:
        200:
            description: The count of sessions
            schema:
                type: integer
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        kwargs = request.args.to_dict()
        count = session_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception as e:
        return jsonify({"message": str(e)}), 500


@bp.route("/count", methods=["POST"])
@jwt_auth_required(role="admin")
def sessions_count():
    """Sessions count
    Retrieves the count of sessions based on the provided parameters.
    Works much like the GET /sessions/count endpoint, but allows for more complex queries.
    By specifying a parameter in the request body, you can filter the sessions based on that parameter,
    if the parameter value contains a comma, the filter will be an "in" query, meaning that the sessions
    will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
        - Sessions
    parameters:
      - name: session
        in: body
        required: false
        schema:
          type: object
          properties:
            status:
              type: string
            session_id:
              type: string
    responses:
        200:
            description: The count of sessions
            schema:
                type: integer
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        kwargs = get_post_data_to_kwargs(request)
        count = session_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception as e:
        return jsonify({"message": str(e)}), 500


@bp.route("/<string:id>", methods=["GET"])
@jwt_auth_required(role="admin")
def get_session(id: str):
    """Get session
    Retrieves a session based on the provided id.
    ---
    tags:
        - Sessions
    parameters:
      - name: id
        in: path
        required: true
        type: string
        description: The id of the session
    responses:
        200:
            description: The session
            schema:
                $ref: '#/definitions/Session'
        404:
            description: The session was not found
            schema:
                type: object
                properties:
                    error:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        session = session_store.get(id, use_typing=False)
        response = session

        return jsonify(response), 200
    except EntityNotFound as e:
        return jsonify({"message": str(e)}), 404
    except Exception as e:
        return jsonify({"message": str(e)}), 500


@bp.route("/", methods=["POST"])
@jwt_auth_required(role="admin")
def post():
    """Create session
    Creates a new session based on the provided data.
    ---
    tags:
        - Sessions
    parameters:
      - name: session
        in: body
        required: true
        schema:
          type: object
          properties:
            session_id:
              type: string
            session_config:
              type: object
    responses:
        201:
            description: The created session
            schema:
                $ref: '#/definitions/Session'
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        data = request.json if request.headers["Content-Type"] == "application/json" else request.form.to_dict()
        successful, result = session_store.add(data)
        response = result
        status_code: int = 201 if successful else 400

        return jsonify(response), status_code
    except Exception as e:
        return jsonify({"message": str(e)}), 500


@bp.route("/start", methods=["POST"])
@jwt_auth_required(role="admin")
def start_session():
    """Start a new session.
    param: session_id: The session id to start.
    type: session_id: str
    param: rounds: The number of rounds to run.
    type: rounds: int
    """
    try:
        data = request.json if request.headers["Content-Type"] == "application/json" else request.form.to_dict()
        session_id: str = data.get("session_id")
        rounds: int = data.get("rounds", "")

        if not session_id or session_id == "":
            return jsonify({"message": "Session ID is required"}), 400

        if not rounds or rounds == "":
            return jsonify({"message": "Rounds is required"}), 400

        if not isinstance(rounds, int):
            return jsonify({"message": "Rounds must be an integer"}), 400

        session = session_store.get(session_id, use_typing=False)

        session_config = session["session_config"]
        model_id = session_config["model_id"]

        _ = model_store.get(model_id, use_typing=False)

        threading.Thread(target=control.start_session, args=(session_id, rounds)).start()

        return jsonify({"message": "Session started"}), 200
    except Exception as e:
        return jsonify({"message": str(e)}), 500



from fedn.common.exceptions import InvalidParameterError


class Parameters(dict):
    """Represents a collection of parameters.

    Extends dict and adds functionality to validate
    paramteres types against a user-provided schema.

    Example of use:
        p = Parameters({'n_iter': 10, 'beta': 1e-2})
        p.validate({'n_iter': int, 'beta': float})

    """

    def __init__(self, parameters=None):
        """ """
        if parameters:
            for key, value in parameters.items():
                self.__setitem__(key, value)

    def validate(self, parameter_schema):
        """Validate parameters against a schema.

        :param parameter_schema: mapping of parameter name and data type.
        :type parameter_schema: dict
        :return: True if the parameters validate
        :rtype: bool
        """
        for key, value in self.items():
            if key not in parameter_schema.keys():
                raise InvalidParameterError("Parameter {} not in paramter schema".format(key))
            else:
                type = parameter_schema[key]
                self._validate_parameter_type(key, value, type)

        return True

    def _validate_parameter_type(self, key, value, type):
        """Validate that parameters values matches the data type.

        :param key: mapping of parameter name and data type.
        :type parameter_schema: dict
        :return: True if the parameters validate
        :rtype: bool

        """
        if not isinstance(value, type):
            raise InvalidParameterError("Parameter {} has invalid type, expecting {}.".format(key, type))

        return True



from abc import ABC


class Config(ABC):
    pass


class ReducerConfig(Config):
    """Configuration for the Reducer component."""

    compute_bundle_dir = None
    models_dir = None

    initial_model = None

    storage_backend = {"type": "s3", "settings": {"bucket": "models"}}

    def __init__(self):
        pass



import abc


class RepositoryBase(object):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def set_artifact(self, instance_name, instance, bucket):
        """Set object with name object_name

        :param instance_name: The name of the object
        :tyep insance_name: str
        :param instance: the object
        :param bucket: The bucket name
        :type bucket: str
        """
        raise NotImplementedError("Must be implemented by subclass")

    @abc.abstractmethod
    def get_artifact(self, instance_name, bucket):
        """Retrive object with name instance_name.

        :param instance_name: The name of the object to retrieve
        :type instance_name: str
        :param bucket: The bucket name
        :type bucket: str
        """
        raise NotImplementedError("Must be implemented by subclass")

    @abc.abstractmethod
    def get_artifact_stream(self, instance_name, bucket):
        """Return a stream handler for object with name instance_name.

        :param instance_name: The name if the object
        :type instance_name: str
        :param bucket: The bucket name
        :type bucket: str
        :return: stream handler for object instance name
        """
        raise NotImplementedError("Must be implemented by subclass")



from abc import ABC, abstractmethod


class StateStoreBase(ABC):
    """ """

    def __init__(self):
        pass

    @abstractmethod
    def state(self):
        """Return the current state of the statestore."""
        pass

    @abstractmethod
    def transition(self, state):
        """Transition the statestore to a new state.

        :param state: The new state.
        :type state: str
        """
        pass

    @abstractmethod
    def set_latest_model(self, model_id):
        """Set the latest model id in the statestore.

        :param model_id: The model id.
        :type model_id: str
        """
        pass

    @abstractmethod
    def get_latest_model(self):
        """Get the latest model id from the statestore.

        :return: The model object.
        :rtype: ObjectId
        """
        pass

    @abstractmethod
    def is_inited(self):
        """Check if the statestore is initialized.

        :return: True if initialized, else False.
        :rtype: bool
        """
        pass



import logging
import logging.config
import os

import requests
import urllib3

log_levels = {"DEBUG": logging.DEBUG, "INFO": logging.INFO, "WARNING": logging.WARNING, "ERROR": logging.ERROR, "CRITICAL": logging.CRITICAL}


urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)
logging.getLogger("urllib3").setLevel(logging.ERROR)

handler = logging.StreamHandler()
logger = logging.getLogger("fedn")
logger.addHandler(handler)
logger.setLevel(logging.DEBUG)
formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(message)s", datefmt="%Y-%m-%d %H:%M:%S")
handler.setFormatter(formatter)


class StudioHTTPHandler(logging.handlers.HTTPHandler):
    def __init__(self, host, url, method="POST", token=None):
        super().__init__(host, url, method)
        self.token = token

    def emit(self, record):
        log_entry = self.mapLogRecord(record)

        log_entry = {
            "msg": log_entry["msg"],
            "levelname": log_entry["levelname"],
            "project": os.environ.get("PROJECT_ID"),
            "appinstance": os.environ.get("APP_ID"),
        }
        # Setup headers
        headers = {
            "Content-type": "application/json",
        }
        if self.token:
            remote_token_protocol = os.environ.get("FEDN_REMOTE_LOG_TOKEN_PROTOCOL", "Token")
            headers["Authorization"] = f"{remote_token_protocol} {self.token}"
        if self.method.lower() == "post":
            requests.post(self.host + self.url, json=log_entry, headers=headers)
        else:
            # No other methods implemented.
            return


# Remote logging can only be configured via environment variables for now.
REMOTE_LOG_SERVER = os.environ.get("FEDN_REMOTE_LOG_SERVER", False)
REMOTE_LOG_PATH = os.environ.get("FEDN_REMOTE_LOG_PATH", False)
REMOTE_LOG_LEVEL = os.environ.get("FEDN_REMOTE_LOG_LEVEL", "INFO")

if REMOTE_LOG_SERVER:
    rloglevel = log_levels.get(REMOTE_LOG_LEVEL, logging.INFO)
    remote_token = os.environ.get("FEDN_REMOTE_LOG_TOKEN", None)

    http_handler = StudioHTTPHandler(host=REMOTE_LOG_SERVER, url=REMOTE_LOG_PATH, method="POST", token=remote_token)
    http_handler.setLevel(rloglevel)
    logger.addHandler(http_handler)


def set_log_level_from_string(level_str):
    """Set the log level based on a string input.
    """
    # Mapping of string representation to logging constants
    level_mapping = {
        "CRITICAL": logging.CRITICAL,
        "ERROR": logging.ERROR,
        "WARNING": logging.WARNING,
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
    }

    # Get the logging level from the mapping
    level = level_mapping.get(level_str.upper())

    if not level:
        raise ValueError(f"Invalid log level: {level_str}")

    # Set the log level
    logger.setLevel(level)


def set_log_stream(log_file):
    """Redirect the log stream to a specified file, if log_file is set.
    """
    if not log_file:
        return

    # Remove existing handlers
    for h in logger.handlers[:]:
        logger.removeHandler(h)

    # Create a FileHandler
    file_handler = logging.FileHandler(log_file)
    file_handler.setFormatter(formatter)

    # Add the file handler to the logger
    logger.addHandler(file_handler)



import uuid

import click
import requests

from fedn.network.combiner.combiner import Combiner

from .main import main
from .shared import CONTROLLER_DEFAULTS, apply_config, get_api_url, get_token, print_response


@main.group("combiner")
@click.pass_context
def combiner_cmd(ctx):
    """:param ctx:
    """
    pass


@combiner_cmd.command("start")
@click.option("-d", "--discoverhost", required=False, help="Hostname for discovery services (reducer).")
@click.option("-p", "--discoverport", required=False, help="Port for discovery services (reducer).")
@click.option("-t", "--token", required=False, help="Set token provided by reducer if enabled")
@click.option("-n", "--name", required=False, default="combiner" + str(uuid.uuid4())[:8], help="Set name for combiner.")
@click.option("-h", "--host", required=False, default="combiner", help="Set hostname.")
@click.option("-i", "--port", required=False, default=12080, help="Set port.")
@click.option("-f", "--fqdn", required=False, default=None, help="Set fully qualified domain name")
@click.option("-s", "--secure", is_flag=True, help="Enable SSL/TLS encrypted gRPC channels.")
@click.option("-v", "--verify", is_flag=True, help="Verify SSL/TLS for REST discovery service (reducer)")
@click.option("-c", "--max_clients", required=False, default=30, help="The maximal number of client connections allowed.")
@click.option("-in", "--init", required=False, default=None, help="Path to configuration file to (re)init combiner.")
@click.pass_context
def start_cmd(ctx, discoverhost, discoverport, token, name, host, port, fqdn, secure, verify, max_clients, init):
    """:param ctx:
    :param discoverhost:
    :param discoverport:
    :param token:
    :param name:
    :param hostname:
    :param port:
    :param secure:
    :param max_clients:
    :param init:
    """
    config = {
        "discover_host": discoverhost,
        "discover_port": discoverport,
        "token": token,
        "host": host,
        "port": port,
        "fqdn": fqdn,
        "name": name,
        "secure": secure,
        "verify": verify,
        "max_clients": max_clients,
    }

    if init:
        apply_config(init, config)
        click.echo(f"\nCombiner configuration loaded from file: {init}")
        click.echo("Values set in file override defaults and command line arguments...\n")

    combiner = Combiner(config)
    combiner.run()


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("--n_max", required=False, help="Number of items to list")
@combiner_cmd.command("list")
@click.pass_context
def list_combiners(ctx, protocol: str, host: str, port: str, token: str = None, n_max: int = None):
    """Return:
    ------
    - count: number of combiners
    - result: list of combiners

    """
    url = get_api_url(protocol=protocol, host=host, port=port, endpoint="combiners")
    headers = {}

    if n_max:
        headers["X-Limit"] = n_max

    _token = get_token(token)

    if _token:
        headers["Authorization"] = _token

    click.echo(f"\nListing combiners: {url}\n")
    click.echo(f"Headers: {headers}")

    try:
        response = requests.get(url, headers=headers)
        print_response(response, "combiners")
    except requests.exceptions.ConnectionError:
        click.echo(f"Error: Could not connect to {url}")



import uuid

from fedn.common.log_config import logger
from fedn.network.storage.s3.miniorepository import MINIORepository


class Repository:
    """Interface for storing model objects and compute packages in S3 compatible storage."""

    def __init__(self, config):
        self.model_bucket = config["storage_bucket"]
        self.context_bucket = config["context_bucket"]

        # TODO: Make a plug-in solution
        self.client = MINIORepository(config)

        self.client.create_bucket(self.context_bucket)
        self.client.create_bucket(self.model_bucket)

    def get_model(self, model_id):
        """Retrieve a model with id model_id.

        :param model_id: Unique identifier for model to retrive.
        :return: The model object
        """
        logger.info("Client {} trying to get model with id: {}".format(self.client.name, model_id))
        return self.client.get_artifact(model_id, self.model_bucket)

    def get_model_stream(self, model_id):
        """Retrieve a stream handle to model with id model_id.

        :param model_id:
        :return: Handle to model object
        """
        logger.info("Client {} trying to get model with id: {}".format(self.client.name, model_id))
        return self.client.get_artifact_stream(model_id, self.model_bucket)

    def set_model(self, model, is_file=True):
        """Upload model object.

        :param model: The model object
        :type model: BytesIO or str file name.
        :param is_file: True if model is a file name, else False
        :return: id for the uploaded object (str)
        """
        model_id = uuid.uuid4()

        try:
            self.client.set_artifact(str(model_id), model, bucket=self.model_bucket, is_file=is_file)
        except Exception:
            logger.error("Failed to upload model with ID {} to repository.".format(model_id))
            raise
        return str(model_id)

    def delete_model(self, model_id):
        """Delete model.

        :param model_id: The id of the model to delete
        :type model_id: str
        """
        try:
            self.client.delete_artifact(model_id, bucket=self.model_bucket)
        except Exception:
            logger.error("Failed to delete model {} repository.".format(model_id))
            raise

    def set_compute_package(self, name, compute_package, is_file=True):
        """Upload compute package.

        :param name: The name of the compute package.
        :type name: str
        :param compute_package: The compute package
        :type compute_pacakge: BytesIO or str file name.
        :param is_file: True if model is a file name, else False
        """
        try:
            self.client.set_artifact(str(name), compute_package, bucket=self.context_bucket, is_file=is_file)
        except Exception:
            logger.error("Failed to write compute_package to repository.")
            raise

    def get_compute_package(self, compute_package):
        """Retrieve compute package from object store.

        :param compute_package: The name of the compute package.
        :type compute_pacakge: str
        :return: Compute package.
        """
        try:
            data = self.client.get_artifact(compute_package, bucket=self.context_bucket)
        except Exception:
            logger.error("Failed to get compute_package from repository.")
            raise
        return data

    def delete_compute_package(self, compute_package):
        """Delete a compute package from storage.

        :param compute_package: The name of the compute_package
        :type compute_package: str
        """
        try:
            self.client.delete_artifact(compute_package, bucket=[self.context_bucket])
        except Exception:
            logger.error("Failed to delete compute_package from repository.")
            raise



import os
import unittest

from fedn.utils.helpers.helpers import get_helper, save_metadata, save_metrics


class TestHelpers(unittest.TestCase):

    def test_get_helper(self):
        helper = get_helper('pytorchhelper')

        # Check that helper is not None
        self.assertTrue(helper is not None)

        # Check that helper nane is correct
        self.assertTrue(helper.name == 'pytorchhelper')

    def test_save_metadata(self):
        metadata = {'test': 'test'}
        save_metadata(metadata, 'test')

        # Check that file exists
        self.assertTrue(os.path.exists('test-metadata'))

        # Check that file is not empty
        self.assertTrue(os.path.getsize('test-metadata') > 0)

        # Check that file contains the correct data
        with open('test-metadata', 'r') as f:
            data = f.read()
            self.assertTrue(data == '{"test": "test"}')

    def test_save_metrics(self):
        metrics = {'test': 'test'}
        save_metrics(metrics, 'test_metrics.json')

        # Check that file exists
        self.assertTrue(os.path.exists('test_metrics.json'))

        # Check that file is not empty
        self.assertTrue(os.path.getsize('test_metrics.json') > 0)

        # Check that file contains the correct data
        with open('test_metrics.json', 'r') as f:
            data = f.read()
            self.assertTrue(data == '{"test": "test"}')

    # Clean up (remove files)
    def tearDown(self):
        if os.path.exists('test-metadata'):
            os.remove('test-metadata')
        if os.path.exists('test_metrics.json'):
            os.remove('test_metrics.json')


if __name__ == '__main__':
    unittest.main()



import base64
import json
import queue
import re
import signal
import sys
import threading
import time
import uuid
from datetime import datetime, timedelta
from enum import Enum

import fedn.network.grpc.fedn_pb2 as fedn
import fedn.network.grpc.fedn_pb2_grpc as rpc
from fedn.common.log_config import logger, set_log_level_from_string, set_log_stream
from fedn.network.combiner.connect import ConnectorCombiner, Status
from fedn.network.combiner.modelservice import ModelService
from fedn.network.combiner.roundhandler import RoundHandler
from fedn.network.grpc.server import Server
from fedn.network.storage.s3.repository import Repository
from fedn.network.storage.statestore.mongostatestore import MongoStateStore

VALID_NAME_REGEX = "^[a-zA-Z0-9_-]*$"


class Role(Enum):
    """Enum for combiner roles."""

    WORKER = 1
    COMBINER = 2
    REDUCER = 3
    OTHER = 4


def role_to_proto_role(role):
    """Convert a Role to a proto Role.

    :param role: the role to convert
    :type role: :class:`fedn.network.combiner.server.Role`
    :return: proto role
    :rtype: :class:`fedn.network.grpc.fedn_pb2.Role`
    """
    if role == Role.COMBINER:
        return fedn.COMBINER
    if role == Role.WORKER:
        return fedn.WORKER
    if role == Role.REDUCER:
        return fedn.REDUCER
    if role == Role.OTHER:
        return fedn.OTHER


class Combiner(rpc.CombinerServicer, rpc.ReducerServicer, rpc.ConnectorServicer, rpc.ControlServicer):
    """Combiner gRPC server.

    :param config: configuration for the combiner
    :type config: dict
    """

    def __init__(self, config):
        """Initialize Combiner server."""
        set_log_level_from_string(config.get("verbosity", "INFO"))
        set_log_stream(config.get("logfile", None))

        # Client queues
        self.clients = {}


        # Validate combiner name
        match = re.search(VALID_NAME_REGEX, config["name"])
        if not match:
            raise ValueError("Unallowed character in combiner name. Allowed characters: a-z, A-Z, 0-9, _, -.")

        self.id = config["name"]
        self.role = Role.COMBINER
        self.max_clients = config["max_clients"]

        # Connector to announce combiner to discover service (reducer)
        announce_client = ConnectorCombiner(
            host=config["discover_host"],
            port=config["discover_port"],
            myhost=config["host"],
            fqdn=config["fqdn"],
            myport=config["port"],
            token=config["token"],
            name=config["name"],
            secure=config["secure"],
            verify=config["verify"],
        )

        while True:
            # Announce combiner to discover service
            status, response = announce_client.announce()
            if status == Status.TryAgain:
                logger.info(response)
                time.sleep(5)
            elif status == Status.Assigned:
                announce_config = response
                logger.info("COMBINER {0}: Announced successfully".format(self.id))
                break
            elif status == Status.UnAuthorized:
                logger.info(response)
                logger.info("Status.UnAuthorized")
                sys.exit("Exiting: Unauthorized")
            elif status == Status.UnMatchedConfig:
                logger.info(response)
                logger.info("Status.UnMatchedConfig")
                sys.exit("Exiting: Missing config")

        cert = announce_config["certificate"]
        key = announce_config["key"]

        if announce_config["certificate"]:
            cert = base64.b64decode(announce_config["certificate"])  # .decode('utf-8')
            key = base64.b64decode(announce_config["key"])  # .decode('utf-8')

        # Set up gRPC server configuration
        grpc_config = {"port": config["port"], "secure": config["secure"], "certificate": cert, "key": key}

        # Set up model repository
        self.repository = Repository(announce_config["storage"]["storage_config"])

        self.statestore = MongoStateStore(announce_config["statestore"]["network_id"], announce_config["statestore"]["mongo_config"])

        # Fetch all clients previously connected to the combiner
        # If a client and a combiner goes down at the same time,
        # the client will be stuck listed as "online" in the statestore.
        # Set the status to offline for previous clients.
        previous_clients = self.statestore.clients.find({"combiner": config["name"]})
        for client in previous_clients:
            self.statestore.set_client({"name": client["name"], "status": "offline"})

        self.modelservice = ModelService()

        # Create gRPC server
        self.server = Server(self, self.modelservice, grpc_config)

        # Set up round controller
        self.round_handler = RoundHandler(self.repository, self, self.modelservice)

        # Start thread for round controller
        threading.Thread(target=self.round_handler.run, daemon=True).start()

        # Start thread for client status updates: TODO: Should be configurable
        threading.Thread(target=self._deamon_thread_client_status, daemon=True).start()

        # Start the gRPC server
        self.server.start()

    def __whoami(self, client, instance):
        """Set the client id and role in a proto message.

        :param client: the client to set the id and role for
        :type client: :class:`fedn.network.grpc.fedn_pb2.Client`
        :param instance: the instance to get the id and role from
        :type instance: :class:`fedn.network.combiner.server.Combiner`
        :return: the client with id and role set
        :rtype: :class:`fedn.network.grpc.fedn_pb2.Client`
        """
        client.name = instance.id
        client.role = role_to_proto_role(instance.role)
        return client

    def request_model_update(self, config, clients=[]):
        """Ask clients to update the current global model.

        :param config: the model configuration to send to clients
        :type config: dict
        :param clients: the clients to send the request to
        :type clients: list

        """
        # The request to be added to the client queue
        request = fedn.TaskRequest()
        request.model_id = config["model_id"]
        request.correlation_id = str(uuid.uuid4())
        request.timestamp = str(datetime.now())
        request.data = json.dumps(config)
        request.type = fedn.StatusType.MODEL_UPDATE
        request.session_id = config["session_id"]

        request.sender.name = self.id
        request.sender.role = fedn.COMBINER

        if len(clients) == 0:
            clients = self.get_active_trainers()

        for client in clients:
            request.receiver.name = client
            request.receiver.role = fedn.WORKER
            self._put_request_to_client_queue(request, fedn.Queue.TASK_QUEUE)

        if len(clients) < 20:
            logger.info("Sent model update request for model {} to clients {}".format(request.model_id, clients))
        else:
            logger.info("Sent model update request for model {} to {} clients".format(request.model_id, len(clients)))

    def request_model_validation(self, model_id, config, clients=[]):
        """Ask clients to validate the current global model.

        :param model_id: the model id to validate
        :type model_id: str
        :param config: the model configuration to send to clients
        :type config: dict
        :param clients: the clients to send the request to
        :type clients: list

        """
        # The request to be added to the client queue
        request = fedn.TaskRequest()
        request.model_id = model_id
        request.correlation_id = str(uuid.uuid4())
        request.timestamp = str(datetime.now())
        # request.is_inference = (config['task'] == 'inference')
        request.type = fedn.StatusType.MODEL_VALIDATION

        request.sender.name = self.id
        request.sender.role = fedn.COMBINER
        request.session_id = config["session_id"]

        if len(clients) == 0:
            clients = self.get_active_validators()

        for client in clients:
            request.receiver.name = client
            request.receiver.role = fedn.WORKER
            self._put_request_to_client_queue(request, fedn.Queue.TASK_QUEUE)

        if len(clients) < 20:
            logger.info("Sent model validation request for model {} to clients {}".format(request.model_id, clients))
        else:
            logger.info("Sent model validation request for model {} to {} clients".format(request.model_id, len(clients)))

    def get_active_trainers(self):
        """Get a list of active trainers.

        :return: the list of active trainers
        :rtype: list
        """
        trainers = self._list_active_clients(fedn.Queue.TASK_QUEUE)
        return trainers

    def get_active_validators(self):
        """Get a list of active validators.

        :return: the list of active validators
        :rtype: list
        """
        validators = self._list_active_clients(fedn.Queue.TASK_QUEUE)
        return validators

    def nr_active_trainers(self):
        """Get the number of active trainers.

        :return: the number of active trainers
        :rtype: int
        """
        return len(self.get_active_trainers())

    ####################################################################################################################

    def __join_client(self, client):
        """Add a client to the list of active clients.

        :param client: the client to add
        :type client: :class:`fedn.network.grpc.fedn_pb2.Client`
        """
        if client.name not in self.clients.keys():
            # The status is set to offline by default, and will be updated once _list_active_clients is called.
            self.clients[client.name] = {"lastseen": datetime.now(), "status": "offline"}

    def _subscribe_client_to_queue(self, client, queue_name):
        """Subscribe a client to the queue.

        :param client: the client to subscribe
        :type client: :class:`fedn.network.grpc.fedn_pb2.Client`
        :param queue_name: the name of the queue to subscribe to
        :type queue_name: str
        """
        self.__join_client(client)
        if queue_name not in self.clients[client.name].keys():
            self.clients[client.name][queue_name] = queue.Queue()

    def __get_queue(self, client, queue_name):
        """Get the queue for a client.

        :param client: the client to get the queue for
        :type client: :class:`fedn.network.grpc.fedn_pb2.Client`
        :param queue_name: the name of the queue to get
        :type queue_name: str
        :return: the queue
        :rtype: :class:`queue.Queue`

        :raises KeyError: if the queue does not exist
        """
        try:
            return self.clients[client.name][queue_name]
        except KeyError:
            raise

    def _list_subscribed_clients(self, queue_name):
        """List all clients subscribed to a queue.

        :param queue_name: the name of the queue
        :type queue_name: str
        :return: a list of client names
        :rtype: list
        """
        subscribed_clients = []
        for name, client in self.clients.items():
            if queue_name in client.keys():
                subscribed_clients.append(name)
        return subscribed_clients

    def _list_active_clients(self, channel):
        """List all clients that have sent a status message in the last 10 seconds.

        :param channel: the name of the channel
        :type channel: str
        :return: a list of client names
        :rtype: list
        """
        # Temporary dict to store client status
        clients = {
            "active_clients": [],
            "update_active_clients": [],
            "update_offline_clients": [],
        }
        for client in self._list_subscribed_clients(channel):
            status = self.clients[client]["status"]
            now = datetime.now()
            then = self.clients[client]["lastseen"]
            if (now - then) < timedelta(seconds=10):
                clients["active_clients"].append(client)
                # If client has changed status, update statestore
                if status != "online":
                    self.clients[client]["status"] = "online"
                    clients["update_active_clients"].append(client)
            elif status != "offline":
                self.clients[client]["status"] = "offline"
                clients["update_offline_clients"].append(client)
        # Update statestore with client status
        if len(clients["update_active_clients"]) > 0:
            self.statestore.update_client_status(clients["update_active_clients"], "online")
        if len(clients["update_offline_clients"]) > 0:
            self.statestore.update_client_status(clients["update_offline_clients"], "offline")

        return clients["active_clients"]

    def _deamon_thread_client_status(self, timeout=5):
        """Deamon thread that checks for inactive clients and updates statestore."""
        while True:
            time.sleep(timeout)
            # TODO: Also update validation clients
            self._list_active_clients(fedn.Queue.TASK_QUEUE)

    def _put_request_to_client_queue(self, request, queue_name):
        """Get a client specific queue and add a request to it.
        The client is identified by the request.receiver.

        :param request: the request to send
        :type request: :class:`fedn.network.grpc.fedn_pb2.Request`
        :param queue_name: the name of the queue to send the request to
        :type queue_name: str
        """
        try:
            q = self.__get_queue(request.receiver, queue_name)
            q.put(request)
        except Exception as e:
            logger.error("Failed to put request to client queue {} for client {}: {}".format(queue_name, request.receiver.name, str(e)))
            raise

    def _send_status(self, status):
        """Report a status to backend db.

        :param status: the status to report
        :type status: :class:`fedn.network.grpc.fedn_pb2.Status`
        """
        self.statestore.report_status(status)

    def _flush_model_update_queue(self):
        """Clear the model update queue (aggregator).

        :return: True if successful, else False
        """
        q = self.round_handler.aggregator.model_updates
        try:
            with q.mutex:
                q.queue.clear()
                q.all_tasks_done.notify_all()
                q.unfinished_tasks = 0
            return True
        except Exception as e:
            logger.error("Failed to flush model update queue: %s", str(e))
            return False

    #####################################################################################################################

    # Controller Service

    def Start(self, control: fedn.ControlRequest, context):
        """Start a round of federated learning"

        :param control: the control request
        :type control: :class:`fedn.network.grpc.fedn_pb2.ControlRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the control response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ControlResponse`
        """
        logger.info("grpc.Combiner.Start: Starting round")

        config = {}
        for parameter in control.parameter:
            config.update({parameter.key: parameter.value})

        logger.debug("grpc.Combiner.Start: Round config {}".format(config))

        job_id = self.round_handler.push_round_config(config)
        logger.info("grcp.Combiner.Start: Pushed round config (job_id): {}".format(job_id))

        response = fedn.ControlResponse()
        p = response.parameter.add()
        p.key = "job_id"
        p.value = job_id

        return response

    def SetAggregator(self, control: fedn.ControlRequest, context):
        """Set the active aggregator.

        :param control: the control request
        :type control: :class:`fedn.network.grpc.fedn_pb2.ControlRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the control response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ControlResponse`
        """
        logger.debug("grpc.Combiner.SetAggregator: Called")
        for parameter in control.parameter:
            aggregator = parameter.value

        status = self.round_handler.set_aggregator(aggregator)

        response = fedn.ControlResponse()
        if status:
            response.message = "Success"
        else:
            response.message = "Failed"

        return response

    def FlushAggregationQueue(self, control: fedn.ControlRequest, context):
        """Flush the queue.

        :param control: the control request
        :type control: :class:`fedn.network.grpc.fedn_pb2.ControlRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the control response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ControlResponse`
        """
        logger.debug("grpc.Combiner.FlushAggregationQueue: Called")
        status = self._flush_model_update_queue()

        response = fedn.ControlResponse()
        if status:
            response.message = "Success"
        else:
            response.message = "Failed"

        return response

    ##############################################################################

    def Stop(self, control: fedn.ControlRequest, context):
        """TODO: Not yet implemented.

        :param control: the control request
        :type control: :class:`fedn.network.grpc.fedn_pb2.ControlRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the control response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ControlResponse`
        """
        response = fedn.ControlResponse()
        logger.info("grpc.Combiner.Stop: Called")
        return response

    #####################################################################################################################

    def SendStatus(self, status: fedn.Status, context):
        """A client RPC endpoint that accepts status messages.

        :param status: the status message
        :type status: :class:`fedn.network.grpc.fedn_pb2.Status`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.Response`
        """
        logger.debug("grpc.Combiner.SendStatus: Called")
        self._send_status(status)

        response = fedn.Response()
        response.response = "Status received."
        return response

    def ListActiveClients(self, request: fedn.ListClientsRequest, context):
        """RPC endpoint that returns a ClientList containing the names of all active clients.
            An active client has sent a status message / responded to a heartbeat
            request in the last 10 seconds.

        :param request: the request
        :type request: :class:`fedn.network.grpc.fedn_pb2.ListClientsRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the client list
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ClientList`
        """
        clients = fedn.ClientList()
        active_clients = self._list_active_clients(request.channel)
        nr_active_clients = len(active_clients)
        if nr_active_clients < 20:
            logger.info("grpc.Combiner.ListActiveClients:  Active clients: {}".format(active_clients))
        else:
            logger.info("grpc.Combiner.ListActiveClients: Number active clients: {}".format(nr_active_clients))

        for client in active_clients:
            clients.client.append(fedn.Client(name=client, role=fedn.WORKER))
        return clients

    def AcceptingClients(self, request: fedn.ConnectionRequest, context):
        """RPC endpoint that returns a ConnectionResponse indicating whether the server
        is accepting clients or not.

        :param request: the request (unused)
        :type request: :class:`fedn.network.grpc.fedn_pb2.ConnectionRequest`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.ConnectionResponse`
        """
        response = fedn.ConnectionResponse()
        active_clients = self._list_active_clients(fedn.Queue.TASK_QUEUE)

        try:
            requested = int(self.max_clients)
            if len(active_clients) >= requested:
                response.status = fedn.ConnectionStatus.NOT_ACCEPTING
                return response
            if len(active_clients) < requested:
                response.status = fedn.ConnectionStatus.ACCEPTING
                return response

        except Exception as e:
            logger.error("Combiner not properly configured! {}".format(e))
            raise

        response.status = fedn.ConnectionStatus.TRY_AGAIN_LATER
        return response

    def SendHeartbeat(self, heartbeat: fedn.Heartbeat, context):
        """RPC that lets clients send a hearbeat, notifying the server that
            the client is available.

        :param heartbeat: the heartbeat
        :type heartbeat: :class:`fedn.network.grpc.fedn_pb2.Heartbeat`
        :param context: the context (unused)
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.Response`
        """
        logger.debug("GRPC: Received heartbeat from {}".format(heartbeat.sender.name))
        # Update the clients dict with the last seen timestamp.
        client = heartbeat.sender
        self.__join_client(client)
        self.clients[client.name]["lastseen"] = datetime.now()

        response = fedn.Response()
        response.sender.name = heartbeat.sender.name
        response.sender.role = heartbeat.sender.role
        response.response = "Heartbeat received"
        return response

    # Combiner Service

    def TaskStream(self, response, context):
        """A server stream RPC endpoint (Update model). Messages from client stream.

        :param response: the response
        :type response: :class:`fedn.network.grpc.fedn_pb2.ModelUpdateRequest`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        """
        client = response.sender
        metadata = context.invocation_metadata()
        if metadata:
            metadata = dict(metadata)
            logger.info("grpc.Combiner.TaskStream: Client connected: {}\n".format(metadata["client"]))

        status = fedn.Status(status="Client {} connecting to TaskStream.".format(client.name))
        status.log_level = fedn.Status.INFO
        status.timestamp.GetCurrentTime()

        self.__whoami(status.sender, self)

        self._subscribe_client_to_queue(client, fedn.Queue.TASK_QUEUE)
        q = self.__get_queue(client, fedn.Queue.TASK_QUEUE)

        self._send_status(status)

        # Set client status to online
        self.clients[client.name]["status"] = "online"
        self.statestore.set_client({"name": client.name, "status": "online"})

        # Keep track of the time context has been active
        start_time = time.time()
        while context.is_active():
            # Check if the context has been active for more than 10 seconds
            if time.time() - start_time > 10:
                self.clients[client.name]["lastseen"] = datetime.now()
                # Reset the start time
                start_time = time.time()
            try:
                yield q.get(timeout=1.0)
            except queue.Empty:
                pass
            except Exception as e:
                logger.error("Error in ModelUpdateRequestStream: {}".format(e))

    def SendModelUpdate(self, request, context):
        """Send a model update response.

        :param request: the request
        :type request: :class:`fedn.network.grpc.fedn_pb2.ModelUpdate`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.Response`
        """
        self.round_handler.aggregator.on_model_update(request)

        response = fedn.Response()
        response.response = "RECEIVED ModelUpdate {} from client  {}".format(response, response.sender.name)
        return response  # TODO Fill later

    def register_model_validation(self, validation):
        """Register a model validation.

        :param validation: the model validation
        :type validation: :class:`fedn.network.grpc.fedn_pb2.ModelValidation`
        """
        self.statestore.report_validation(validation)

    def SendModelValidation(self, request, context):
        """Send a model validation response.

        :param request: the request
        :type request: :class:`fedn.network.grpc.fedn_pb2.ModelValidation`
        :param context: the context
        :type context: :class:`grpc._server._Context`
        :return: the response
        :rtype: :class:`fedn.network.grpc.fedn_pb2.Response`
        """
        logger.info("Recieved ModelValidation from {}".format(request.sender.name))

        self.register_model_validation(request)

        response = fedn.Response()
        response.response = "RECEIVED ModelValidation {} from client  {}".format(response, response.sender.name)
        return response

    ####################################################################################################################

    def run(self):
        """Start the server."""
        logger.info("COMBINER: {} started, ready for gRPC requests.".format(self.id))
        try:
            while True:
                signal.pause()
        except (KeyboardInterrupt, SystemExit):
            pass
        self.server.stop()



from abc import ABC, abstractmethod


class HelperBase(ABC):
    """Abstract class defining helpers."""

    def __init__(self):
        """Initialize helper."""
        self.name = self.__class__.__name__

    @abstractmethod
    def increment_average(self, m1, m2, a, W):
        """Compute one increment of incremental weighted averaging.

        :param m1: Current model weights in array-like format.
        :param m2: New model weights in array-like format.
        :param a: Number of examples in new model.
        :param W: Total number of examples.
        :return: Incremental weighted average of model weights.
        """
        pass

    @abstractmethod
    def save(self, model, path):
        """Serialize weights to file. The serialized model must be a single binary object.

        :param model: Weights in array-like format.
        :param path: Path to file.

        """
        pass

    @abstractmethod
    def load(self, fh):
        """Load weights from file or filelike.

        :param fh: file path, filehandle, filelike.
        :return: Weights in array-like format.
        """
        pass



from flask import Blueprint, jsonify, request

from fedn.network.api.auth import jwt_auth_required
from fedn.network.api.v1.shared import api_version, get_post_data_to_kwargs, get_typed_list_headers, get_use_typing, mdb
from fedn.network.storage.statestore.stores.shared import EntityNotFound
from fedn.network.storage.statestore.stores.status_store import StatusStore

bp = Blueprint("status", __name__, url_prefix=f"/api/{api_version}/statuses")

status_store = StatusStore(mdb, "control.status")


@bp.route("/", methods=["GET"])
@jwt_auth_required(role="admin")
def get_statuses():
    """Get statuses
    Retrieves a list of statuses based on the provided parameters.
    By specifying a parameter in the url, you can filter the statuses based on that parameter,
    and the response will contain only the statuses that match the filter.
    ---
    tags:
        - Statuses
    parameters:
      - name: type
        in: query
        required: false
        type: string
        description: Type of the status
      - name: sender.name
        in: query
        required: false
        type: string
        description: Name of the sender
      - name: sender.role
        in: query
        required: false
        type: string
        description: Role of the sender
      - name: session_id
        in: query
        required: false
        type: string
      - name: log_level
        in: query
        required: false
        type: string
        description: Log level of the status
      - name: correlation_id
        in: query
        required: false
        type: string
        description: Correlation id of the status
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of statuses to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of statuses to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the statuses by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the statuses in ('asc' or 'desc')
    definitions:
      Status:
        type: object
        properties:
          id:
            type: string
          status:
            type: string
          session_id:
            type: string
          timestamp:
            type: object
            format: date-time
          type:
            type: string
          data:
            type: string
          correlation_id:
            type: string
          extra:
            type: string
          sender:
            type: object
            properties:
                name:
                    type: string
                role:
                    type: string
          log_level:
            type: object
    responses:
      200:
        description: A list of statuses and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Session'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, use_typing = get_typed_list_headers(request.headers)
        kwargs = request.args.to_dict()

        statuses = status_store.list(limit, skip, sort_key, sort_order, use_typing=use_typing, **kwargs)

        result = [status.__dict__ for status in statuses["result"]] if use_typing else statuses["result"]

        response = {"count": statuses["count"], "result": result}

        return jsonify(response), 200
    except Exception as e:
        return jsonify({"message": str(e)}), 500


@bp.route("/list", methods=["POST"])
@jwt_auth_required(role="admin")
def list_statuses():
    """Get statuses
    Retrieves a list of statuses based on the provided parameters.
    Works much like the GET statuses method, but allows for a more complex query.
    By specifying a parameter in the body, you can filter the statuses based on that parameter,
    and the response will contain only the statuses that match the filter. If the parameter value contains a comma,
    the filter will be an "in" query, meaning that the statuses will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
        - Statuses
    parameters:
      - name: status
        in: body
        required: false
        schema:
            type: object
            properties:
                type:
                    type: string
                    description: Type of the status
                sender.name:
                    type: string
                    description: Name of the sender
                sender.role:
                    required: false
                    type: string
                    description: Role of the sender
                session_id:
                    required: false
                    type: string
                log_level:
                    required: false
                    type: string
                    description: Log level of the status
                correlation_id:
                    required: false
                    type: string
                    description: Correlation id of the status
      - name: X-Limit
        in: header
        required: false
        type: integer
        description: The maximum number of statuses to retrieve
      - name: X-Skip
        in: header
        required: false
        type: integer
        description: The number of statuses to skip
      - name: X-Sort-Key
        in: header
        required: false
        type: string
        description: The key to sort the statuses by
      - name: X-Sort-Order
        in: header
        required: false
        type: string
        description: The order to sort the statuses in ('asc' or 'desc')
    responses:
      200:
        description: A list of statuses and the total count.
        schema:
            type: object
            properties:
                count:
                    type: integer
                result:
                    type: array
                    items:
                        $ref: '#/definitions/Session'
      500:
        description: An error occurred
        schema:
            type: object
            properties:
                message:
                    type: string
    """
    try:
        limit, skip, sort_key, sort_order, use_typing = get_typed_list_headers(request.headers)
        kwargs = get_post_data_to_kwargs(request)

        statuses = status_store.list(limit, skip, sort_key, sort_order, use_typing=use_typing, **kwargs)

        result = [status.__dict__ for status in statuses["result"]] if use_typing else statuses["result"]

        response = {"count": statuses["count"], "result": result}

        return jsonify(response), 200
    except Exception as e:
        return jsonify({"message": str(e)}), 500


@bp.route("/count", methods=["GET"])
@jwt_auth_required(role="admin")
def get_statuses_count():
    """Statuses count
    Retrieves the count of statuses based on the provided parameters.
    By specifying a parameter in the url, you can filter the statuses based on that parameter,
    and the response will contain only the count of statuses that match the filter.
    ---
    tags:
        - Statuses
    parameters:
      - name: type
        in: query
        required: false
        type: string
        description: Type of the status
      - name: sender.name
        in: query
        required: false
        type: string
        description: Name of the sender
      - name: sender.role
        in: query
        required: false
        type: string
        description: Role of the sender
      - name: session_id
        in: query
        required: false
        type: string
      - name: log_level
        in: query
        required: false
        type: string
        description: Log level of the status
      - name: correlation_id
        in: query
        required: false
        type: string
        description: Correlation id of the status
    responses:
        200:
            description: The count of statuses
            schema:
                type: integer
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        kwargs = request.args.to_dict()
        count = status_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception as e:
        return jsonify({"message": str(e)}), 500


@bp.route("/count", methods=["POST"])
@jwt_auth_required(role="admin")
def statuses_count():
    """Statuses count
    Retrieves the count of statuses based on the provided parameters.
    Works much like the GET /statuses/count endpoint, but allows for more complex queries.
    By specifying a parameter in the request body, you can filter the statuses based on that parameter,
    if the parameter value contains a comma, the filter will be an "in" query, meaning that the statuses
    will be returned if the specified field contains any of the values in the parameter.
    ---
    tags:
        - Statuses
    parameters:
      - name: status
        in: body
        required: false
        schema:
            type: object
            properties:
                type:
                    type: string
                    description: Type of the status
                sender.name:
                    type: string
                    description: Name of the sender
                sender.role:
                    required: false
                    type: string
                    description: Role of the sender
                session_id:
                    required: false
                    type: string
                log_level:
                    required: false
                    type: string
                    description: Log level of the status
                correlation_id:
                    required: false
                    type: string
                    description: Correlation id of the status
    responses:
        200:
            description: The count of statuses
            schema:
                type: integer
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        kwargs = get_post_data_to_kwargs(request)
        count = status_store.count(**kwargs)
        response = count
        return jsonify(response), 200
    except Exception as e:
        return jsonify({"message": str(e)}), 500


@bp.route("/<string:id>", methods=["GET"])
@jwt_auth_required(role="admin")
def get_status(id: str):
    """Get status
    Retrieves a status based on the provided id.
    ---
    tags:
        - Statuses
    parameters:
      - name: id
        in: path
        required: true
        type: string
        description: The id of the status
    responses:
        200:
            description: The status
            schema:
                $ref: '#/definitions/Status'
        404:
            description: The status was not found
            schema:
                type: object
                properties:
                    error:
                        type: string
        500:
            description: An error occurred
            schema:
                type: object
                properties:
                    message:
                        type: string
    """
    try:
        use_typing: bool = get_use_typing(request.headers)
        status = status_store.get(id, use_typing=use_typing)

        response = status.__dict__ if use_typing else status

        return jsonify(response), 200
    except EntityNotFound as e:
        return jsonify({"message": str(e)}), 404
    except Exception as e:
        return jsonify({"message": str(e)}), 500



from enum import Enum


class ClientState(Enum):
    """Enum for representing the state of a client."""

    idle = 1
    training = 2
    validating = 3


def ClientStateToString(state):
    """Convert a ClientState to a string representation.

    :param state: the state to convert
    :type state: :class:`fedn.network.clients.state.ClientState`
    :return: string representation of the state
    :rtype: str
    """
    if state == ClientState.idle:
        return "IDLE"
    if state == ClientState.training:
        return "TRAINING"
    if state == ClientState.validating:
        return "VALIDATING"

    return "UNKNOWN"



import click
import requests

from .main import main
from .shared import CONTROLLER_DEFAULTS, get_api_url, get_token, print_response


@main.group("status")
@click.pass_context
def status_cmd(ctx):
    """:param ctx:
    """
    pass


@click.option("-p", "--protocol", required=False, default=CONTROLLER_DEFAULTS["protocol"], help="Communication protocol of controller (api)")
@click.option("-H", "--host", required=False, default=CONTROLLER_DEFAULTS["host"], help="Hostname of controller (api)")
@click.option("-P", "--port", required=False, default=CONTROLLER_DEFAULTS["port"], help="Port of controller (api)")
@click.option("-t", "--token", required=False, help="Authentication token")
@click.option("--n_max", required=False, help="Number of items to list")
@status_cmd.command("list")
@click.pass_context
def list_statuses(ctx, protocol: str, host: str, port: str, token: str = None, n_max: int = None):
    """Return:
    ------
    - count: number of statuses
    - result: list of statuses

    """
    url = get_api_url(protocol=protocol, host=host, port=port, endpoint="statuses")
    headers = {}

    if n_max:
        headers["X-Limit"] = n_max

    _token = get_token(token)

    if _token:
        headers["Authorization"] = _token

    click.echo(f"\nListing statuses: {url}\n")
    click.echo(f"Headers: {headers}")

    try:
        response = requests.get(url, headers=headers)
        print_response(response, "statuses")
    except requests.exceptions.ConnectionError:
        click.echo(f"Error: Could not connect to {url}")



""" The controller package is responsible for orchestrating the federated learning process. It's acts as a gRPC client and sends round config tasks 
to the :class:`fedn.network.combiner.Combiner`. """
# flake8: noqa



import json
from datetime import datetime

import numpy
import plotly
import plotly.graph_objs as go
from plotly.subplots import make_subplots

from fedn.common.log_config import logger
from fedn.network.storage.statestore.mongostatestore import MongoStateStore


class Plot:
    """ """

    def __init__(self, statestore):
        try:
            statestore_config = statestore.get_config()
            statestore = MongoStateStore(statestore_config["network_id"], statestore_config["mongo_config"])
            self.mdb = statestore.connect()
            self.status = self.mdb["control.status"]
            self.round_time = self.mdb["control.round_time"]
            self.combiner_round_time = self.mdb["control.combiner_round_time"]
            self.psutil_usage = self.mdb["control.psutil_monitoring"]
            self.network_clients = self.mdb["network.clients"]

        except Exception as e:
            logger.error("FAILED TO CONNECT TO MONGO, {}".format(e))
            self.collection = None
            raise

    # plot metrics from DB
    def _scalar_metrics(self, metrics):
        """Extract all scalar valued metrics from a MODEL_VALIDATON."""
        data = json.loads(metrics["data"])
        data = json.loads(data["data"])

        valid_metrics = []
        for metric, val in data.items():
            # If it can be converted to a float it is a valid, scalar metric
            try:
                val = float(val)
                valid_metrics.append(metric)
            except Exception:
                pass

        return valid_metrics

    def create_table_plot(self):
        """:return:
        """
        metrics = self.status.find_one({"type": "MODEL_VALIDATION"})
        if metrics is None:
            fig = go.Figure(data=[])
            fig.update_layout(title_text="No data currently available for table mean metrics")
            table = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
            return False

        valid_metrics = self._scalar_metrics(metrics)
        if valid_metrics == []:
            fig = go.Figure(data=[])
            fig.update_layout(title_text="No scalar metrics found")
            table = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
            return False

        all_vals = []
        models = []
        for metric in valid_metrics:
            validations = {}
            for post in self.status.find({"type": "MODEL_VALIDATION"}):
                e = json.loads(post["data"])
                try:
                    validations[e["modelId"]].append(float(json.loads(e["data"])[metric]))
                except KeyError:
                    validations[e["modelId"]] = [float(json.loads(e["data"])[metric])]

            vals = []
            models = []
            for model, data in validations.items():
                vals.append(numpy.mean(data))
                models.append(model)
            all_vals.append(vals)

        header_vals = valid_metrics
        models.reverse()
        values = [models]

        for vals in all_vals:
            vals.reverse()
            values.append(vals)

        fig = go.Figure(
            data=[
                go.Table(
                    header=dict(values=["Model ID"] + header_vals, line_color="darkslategray", fill_color="lightskyblue", align="left"),
                    cells=dict(
                        values=values,  # 2nd column
                        line_color="darkslategray",
                        fill_color="lightcyan",
                        align="left",
                    ),
                )
            ]
        )

        fig.update_layout(title_text="Summary: mean metrics")
        table = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return table

    def create_timeline_plot(self):
        """:return:
        """
        trace_data = []
        x = []
        y = []
        base = []
        for p in self.status.find({"type": "MODEL_UPDATE_REQUEST"}):
            e = json.loads(p["data"])
            cid = e["correlationId"]
            for cc in self.status.find({"sender": p["sender"], "type": "MODEL_UPDATE"}):
                da = json.loads(cc["data"])
                if da["correlationId"] == cid:
                    cp = cc

            cd = json.loads(cp["data"])
            tr = datetime.strptime(e["timestamp"], "%Y-%m-%d %H:%M:%S.%f")
            tu = datetime.strptime(cd["timestamp"], "%Y-%m-%d %H:%M:%S.%f")
            ts = tu - tr
            base.append(tr.timestamp())
            x.append(ts.total_seconds() / 60.0)
            y.append(p["sender"]["name"])

        trace_data.append(
            go.Bar(
                x=y,
                y=x,
                marker=dict(color="royalblue"),
                name="Training",
            )
        )

        x = []
        y = []
        base = []
        for p in self.status.find({"type": "MODEL_VALIDATION_REQUEST"}):
            e = json.loads(p["data"])
            cid = e["correlationId"]
            for cc in self.status.find({"sender": p["sender"], "type": "MODEL_VALIDATION"}):
                da = json.loads(cc["data"])
                if da["correlationId"] == cid:
                    cp = cc
            cd = json.loads(cp["data"])
            tr = datetime.strptime(e["timestamp"], "%Y-%m-%d %H:%M:%S.%f")
            tu = datetime.strptime(cd["timestamp"], "%Y-%m-%d %H:%M:%S.%f")
            ts = tu - tr
            base.append(tr.timestamp())
            x.append(ts.total_seconds() / 60.0)
            y.append(p["sender"]["name"])

        trace_data.append(
            go.Bar(
                x=y,
                y=x,
                marker=dict(color="lightskyblue"),
                name="Validation",
            )
        )

        layout = go.Layout(
            barmode="stack",
            showlegend=True,
        )

        fig = go.Figure(data=trace_data, layout=layout)
        fig.update_xaxes(title_text="Alliance/client")
        fig.update_yaxes(title_text="Time (Min)")
        fig.update_layout(title_text="Alliance timeline")
        timeline = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return timeline

    def create_client_training_distribution(self):
        """:return:
        """
        training = []
        for p in self.status.find({"type": "MODEL_UPDATE"}):
            e = json.loads(p["data"])
            meta = json.loads(e["meta"])
            training.append(meta["exec_training"])

        if not training:
            return False
        fig = go.Figure(data=go.Histogram(x=training))
        fig.update_layout(title_text="Client model training time, mean: {}".format(numpy.mean(training)))
        histogram = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return histogram

    def create_client_histogram_plot(self):
        """:return:
        """
        training = []
        for p in self.status.find({"type": "MODEL_UPDATE"}):
            e = json.loads(p["data"])
            meta = json.loads(e["meta"])
            training.append(meta["exec_training"])

        fig = go.Figure()

        fig.update_layout(
            template="simple_white",
            xaxis=dict(title_text="Time (s)"),
            yaxis=dict(title_text="Number of updates"),
            title="Mean client training time: {}".format(numpy.mean(training)),
            # showlegend=True
        )
        if not training:
            return False

        fig.add_trace(go.Histogram(x=training))

        histogram_plot = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return histogram_plot

    def create_client_plot(self):
        """:return:
        """
        processing = []
        upload = []
        download = []
        training = []
        for p in self.status.find({"type": "MODEL_UPDATE"}):
            e = json.loads(p["data"])
            meta = json.loads(e["meta"])
            upload.append(meta["upload_model"])
            download.append(meta["fetch_model"])
            training.append(meta["exec_training"])
            processing.append(meta["processing_time"])

        fig = go.Figure()
        fig.update_layout(template="simple_white", title="Mean client processing time: {}".format(numpy.mean(processing)), showlegend=True)
        if not processing:
            return False
        data = [numpy.mean(training), numpy.mean(upload), numpy.mean(download)]
        labels = ["Training execution", "Model upload (to combiner)", "Model download (from combiner)"]
        fig.add_trace(go.Pie(labels=labels, values=data))

        client_plot = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return client_plot

    def create_combiner_plot(self):
        """:return:
        """
        waiting = []
        aggregation = []
        model_load = []
        combination = []
        for round in self.mdb["control.round"].find():
            try:
                for combiner in round["combiners"]:
                    data = combiner
                    stats = data["local_round"]["1"]
                    ml = stats["aggregation_time"]["time_model_load"]
                    ag = stats["aggregation_time"]["time_model_aggregation"]
                    combination.append(stats["time_combination"])
                    waiting.append(stats["time_combination"] - ml - ag)
                    model_load.append(ml)
                    aggregation.append(ag)
            except Exception:
                pass

        labels = ["Waiting for client updates", "Aggregation", "Loading model updates from disk"]
        val = [numpy.mean(waiting), numpy.mean(aggregation), numpy.mean(model_load)]
        fig = go.Figure()

        fig.update_layout(template="simple_white", title="Mean combiner round time: {}".format(numpy.mean(combination)), showlegend=True)
        if not combination:
            return False
        fig.add_trace(go.Pie(labels=labels, values=val))
        combiner_plot = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return combiner_plot

    def fetch_valid_metrics(self):
        """:return:
        """
        metrics = self.status.find_one({"type": "MODEL_VALIDATION"})
        valid_metrics = self._scalar_metrics(metrics)
        return valid_metrics

    def create_box_plot(self, metric):
        """:param metric:
        :return:
        """
        metrics = self.status.find_one({"type": "MODEL_VALIDATION"})
        if metrics is None:
            fig = go.Figure(data=[])
            fig.update_layout(title_text="No data currently available for metric distribution over  " "participants")
            box = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
            return box

        valid_metrics = self._scalar_metrics(metrics)
        if valid_metrics == []:
            fig = go.Figure(data=[])
            fig.update_layout(title_text="No scalar metrics found")
            box = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
            return box

        validations = {}
        for post in self.status.find({"type": "MODEL_VALIDATION"}):
            e = json.loads(post["data"])
            try:
                validations[e["modelId"]].append(float(json.loads(e["data"])[metric]))
            except KeyError:
                validations[e["modelId"]] = [float(json.loads(e["data"])[metric])]

        # Make sure validations are plotted in chronological order
        model_trail = self.mdb.control.model.find_one({"key": "model_trail"})
        model_trail_ids = model_trail["model"]
        validations_sorted = []
        for model_id in model_trail_ids:
            try:
                validations_sorted.append(validations[model_id])
            except Exception:
                pass

        validations = validations_sorted

        box = go.Figure()

        y = []
        for j, acc in enumerate(validations):
            # x.append(j)
            y.append(numpy.mean([float(i) for i in acc]))
            if len(acc) >= 2:
                box.add_trace(go.Box(y=acc, name=str(j), marker_color="royalblue", showlegend=False, boxpoints=False))
            else:
                box.add_trace(go.Scatter(x=[str(j)], y=[y[j]], showlegend=False))

        rounds = list(range(len(y)))
        box.add_trace(go.Scatter(x=rounds, y=y, name="Mean"))

        box.update_xaxes(title_text="Rounds")
        box.update_yaxes(tickvals=[0.0, 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 1.0])
        box.update_layout(title_text="Metric distribution over clients: {}".format(metric), margin=dict(l=20, r=20, t=45, b=20))
        box = json.dumps(box, cls=plotly.utils.PlotlyJSONEncoder)
        return box

    def create_round_plot(self):
        """:return:
        """
        trace_data = []
        metrics = self.round_time.find_one({"key": "round_time"})
        if metrics is None:
            fig = go.Figure(data=[])
            fig.update_layout(title_text="No data currently available for round time")
            return False

        for post in self.round_time.find({"key": "round_time"}):
            rounds = post["round"]
            traces_data = post["round_time"]

        trace_data.append(go.Scatter(x=rounds, y=traces_data, mode="lines+markers", name="Reducer"))

        for rec in self.combiner_round_time.find({"key": "combiner_round_time"}):
            c_traces_data = rec["round_time"]

        trace_data.append(go.Scatter(x=rounds, y=c_traces_data, mode="lines+markers", name="Combiner"))

        fig = go.Figure(data=trace_data)
        fig.update_xaxes(title_text="Round")
        fig.update_yaxes(title_text="Time (s)")
        fig.update_layout(title_text="Round time")
        round_t = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return round_t

    def create_cpu_plot(self):
        """:return:
        """
        metrics = self.psutil_usage.find_one({"key": "cpu_mem_usage"})
        if metrics is None:
            fig = go.Figure(data=[])
            fig.update_layout(title_text="No data currently available for MEM and CPU usage")
            cpu = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
            return False

        for post in self.psutil_usage.find({"key": "cpu_mem_usage"}):
            cpu = post["cpu"]
            mem = post["mem"]
            ps_time = post["time"]
            round = post["round"]

        # Create figure with secondary y-axis
        fig = make_subplots(specs=[[{"secondary_y": True}]])
        fig.add_trace(go.Scatter(x=ps_time, y=cpu, mode="lines+markers", name="CPU (%)"))

        fig.add_trace(go.Scatter(x=ps_time, y=mem, mode="lines+markers", name="MEM (%)"))

        fig.add_trace(
            go.Scatter(
                x=ps_time,
                y=round,
                mode="lines+markers",
                name="Round",
            ),
            secondary_y=True,
        )

        fig.update_xaxes(title_text="Date Time")
        fig.update_yaxes(title_text="Percentage (%)")
        fig.update_yaxes(title_text="Round", secondary_y=True)
        fig.update_layout(title_text="CPU loads and memory usage")
        cpu = json.dumps(fig, cls=plotly.utils.PlotlyJSONEncoder)
        return cpu
